import numpy as np

from pastml.models import Model
from pastml.models.F81Model import F81Model

JC = 'JC'


class JCModel(F81Model):

    def __init__(self, states, forest_stats, sf=None, tau=0, optimise_tau=False, parameter_file=None, reoptimise=False, **kwargs):
        kwargs['frequency_smoothing'] = False
        F81Model.__init__(self, states=states, forest_stats=forest_stats, sf=sf, tau=tau, optimise_tau=optimise_tau,
                          frequencies=np.ones(len(states), dtype=np.float64) / len(states),
                          reoptimise=reoptimise, parameter_file=parameter_file, **kwargs)
        self._optimise_frequencies = False
        self.name = JC

    def _print_parameters(self):
        """
        Constructs a string representing parameter values (to be used to logging).

        :return: str representing parameter values
        """
        return '{}' \
               '\tfrequencies\tall equal to {:g}\t(fixed)\n'.format(Model._print_parameters(self), 1 / len(self.states))

    def parse_parameters(self, params, reoptimise=False):
        """
        Update this model's values from the input parameters.
        JC model can only have the basic parameters (scaling factor and smoothing factor, see pastml.models.Model).

        :param params: dict {key->value}
        :param reoptimise: whether these model parameters should be treated as starting values (True)
            or as fixed values (False)
        :return: dict with parameter values (same as input)
        """
        # This model sets equal frequencies
        # and hence should only read the basic parameters (scaling and smoothing factors) from the input file
        return Model.parse_parameters(self, params, reoptimise)



import logging
import os
import sys
import warnings
from collections import defaultdict, Counter
from multiprocessing.pool import ThreadPool

import numpy as np
import pandas as pd
from Bio.Phylo import NewickIO, write
from Bio.Phylo.NewickIO import StringIO
from ete3 import Tree

from pastml import col_name2cat, value2list, STATES, METHOD, CHARACTER, get_personalized_feature_name, numeric2datetime, \
    PASTML_VERSION, _set_up_pastml_logger
from pastml.annotation import preannotate_forest, ForestStats
from pastml.file import get_combined_ancestral_state_file, get_named_tree_file, get_pastml_parameter_file, \
    get_pastml_marginal_prob_file, get_pastml_work_dir
from pastml.ml import MARGINAL_PROBABILITIES, is_ml, is_marginal, MPPA, ml_acr, \
    ML_METHODS, MAP, JOINT, ALL, ML, META_ML_METHODS, MARGINAL_ML_METHODS, get_default_ml_method
from pastml.models import MODEL, SCALING_FACTOR, SMOOTHING_FACTOR
from pastml.models.CustomRatesModel import CustomRatesModel, CUSTOM_RATES
from pastml.models.EFTModel import EFTModel, EFT
from pastml.models.F81Model import F81Model, F81
from pastml.models.HKYModel import HKYModel, HKY, HKY_STATES
from pastml.models.JCModel import JCModel, JC
from pastml.models.JTTModel import JTTModel, JTT, JTT_STATES
from pastml.parsimony import is_parsimonious, parsimonious_acr, ACCTRAN, DELTRAN, DOWNPASS, MP_METHODS, MP, \
    get_default_mp_method
from pastml.tree import name_tree, annotate_dates, DATE, read_forest, DATE_CI, resolve_trees, IS_POLYTOMY, \
    unresolve_trees, clear_extra_features, parse_date
from pastml.visualisation import get_formatted_date
from pastml.visualisation.cytoscape_manager import visualize, TIMELINE_SAMPLED, TIMELINE_NODES, TIMELINE_LTT, \
    DIST_TO_ROOT_LABEL, DATE_LABEL
from pastml.visualisation.itol_manager import generate_itol_annotations
from pastml.visualisation.tree_compressor import REASONABLE_NUMBER_OF_TIPS, VERTICAL, HORIZONTAL, TRIM

model2class = {F81: F81Model, JC: JCModel, CUSTOM_RATES: CustomRatesModel, HKY: HKYModel, JTT: JTTModel, EFT: EFTModel}

warnings.filterwarnings("ignore", append=True)

COPY = 'COPY'


def _serialize_acr(args):
    acr_result, work_dir = args
    out_param_file = \
        os.path.join(work_dir,
                     get_pastml_parameter_file(method=acr_result[METHOD],
                                               model=acr_result[MODEL].name if MODEL in acr_result else None,
                                               column=acr_result[CHARACTER]))

    # Not using DataFrames to speed up document writing
    with open(out_param_file, 'w+') as f:
        f.write('parameter\tvalue\n')
        f.write('pastml_version\t{}\n'.format(PASTML_VERSION))
        for name in sorted(acr_result.keys()):
            if name not in [STATES, MARGINAL_PROBABILITIES, METHOD, MODEL]:
                f.write('{}\t{}\n'.format(name, acr_result[name]))
        f.write('{}\t{}\n'.format(METHOD, acr_result[METHOD]))
        if is_ml(acr_result[METHOD]):
            acr_result[MODEL].save_parameters(f)
    logging.getLogger('pastml').debug('Serialized ACR parameters and statistics for {} to {}.'
                                      .format(acr_result[CHARACTER], out_param_file))

    if is_marginal(acr_result[METHOD]):
        out_mp_file = \
            os.path.join(work_dir,
                         get_pastml_marginal_prob_file(method=acr_result[METHOD], model=acr_result[MODEL].name,
                                                       column=acr_result[CHARACTER]))
        acr_result[MARGINAL_PROBABILITIES].to_csv(out_mp_file, sep='\t', index_label='node')
        logging.getLogger('pastml').debug('Serialized marginal probabilities for {} to {}.'
                                          .format(acr_result[CHARACTER], out_mp_file))


def acr(forest, df=None, columns=None, column2states=None, prediction_method=MPPA, model=F81,
        column2parameters=None, column2rates=None,
        force_joint=True, threads=0,
        reoptimise=False, tau=0, resolve_polytomies=False, frequency_smoothing=False):
    """
    Reconstructs ancestral states for the given tree and
    all the characters specified as columns of the given annotation dataframe.

    :param df: dataframe indexed with tree node names
        and containing characters for which ACR should be performed as columns.
    :type df: pandas.DataFrame
    :param forest: tree or list of trees whose ancestral states are to be reconstructed.
    :type forest: ete3.Tree or list(ete3.Tree)
    :param model: (optional, default is F81) model(s) to be used by PASTML,
        can be either one model to be used for all the characters,
        or a list of different models (in the same order as the annotation dataframe columns)
    :type model: str or list(str)
    :param prediction_method: (optional, default is MPPA) ancestral state prediction method(s) to be used by PASTML,
        can be either one method to be used for all the characters,
        or a list of different methods (in the same order as the annotation dataframe columns)
    :type prediction_method: str or list(str)
    :param column2parameters: an optional way to fix some parameters,
        must be in a form {column: {param: value}},
        where param can be a character state (then the value should specify its frequency between 0 and 1),
        or pastml.ml.SCALING_FACTOR (then the value should be the scaling factor for three branches,
        e.g. set to 1 to keep the original branches). Could also be in a form {column: path_to_param_file}.
    :type column2parameters: dict
    :param reoptimise: (False by default) if set to True and the parameters are specified,
        they will be considered as an optimisation starting point instead, and the parameters will be optimised.
    :type reoptimise: bool
    :param force_joint: (optional, default is True) whether the JOINT state should be added to the MPPA prediction
        even when not selected by the Brier score
    :type force_joint: bool
    :param tau: a smoothing factor to apply to branch lengths during likelihood calculation.
        If set to zero (default), zero internal branches will be collapsed instead.
    :type tau: float

    :param threads: (optional, default is 0, which stands for automatic) number of threads PastML can use for parallezation.
        By default, detected automatically based on the system. Note that PastML will at most use as many threads
        as the number of characters (-c option) being analysed plus one.
    :type threads: int

    :return: list of ACR result dictionaries, one per character.
    :rtype: list(dict)
    """

    logger = logging.getLogger('pastml')

    if isinstance(forest, Tree):
        forest = [forest]

    if columns is None:
        if df is None:
            raise ValueError('Either the tree should be preannotated with character values '
                             'and columns and column2states specified, '
                             'or an annotation dataframe provided!')
        columns = df.columns
        column2states = {column: np.array(sorted([_ for _ in df[column].unique() if not pd.isna(_) and '' != _]))
                         for column in columns}
        preannotate_forest(forest, df=df)

    forest_stats = ForestStats(forest)

    logging.getLogger('pastml').debug('\n=============ACR===============================')

    column2parameters = column2parameters if column2parameters else {}
    column2rates = column2rates if column2rates else {}

    prediction_methods = value2list(len(columns), prediction_method, MPPA)
    models = value2list(len(columns), model, F81)

    def get_states(method, model, column):
        initial_states = column2states[column]
        if not is_ml(method) or model not in {HKY, JTT}:
            return initial_states
        states = HKY_STATES if HKY == model else JTT_STATES
        if not set(initial_states) & set(states):
            raise ValueError('The allowed states for model {} are {}, '
                             'but your annotation file specifies {} as states in column {}.'
                             .format(model, ', '.join(states), ', '.join(initial_states), column))
        state_set = set(states)
        for root in forest:
            for n in root.traverse():
                if hasattr(n, column):
                    n.add_feature(column, state_set & getattr(n, column))
        return states

    # If we gonna resolve polytomies we might need to get back to the initial states so let's memorise them
    n2c2states = defaultdict(dict)
    for root in forest:
        for n in root.traverse():
            for c in columns:
                vs = getattr(n, c, set())
                if vs:
                    n2c2states[n][c] = vs

    # method, model, states, params, optimise
    character2settings = {}
    for (character, prediction_method, model) in zip(columns, prediction_methods, models):
        logging.getLogger('pastml') \
            .debug('ACR settings for {}:\n\tMethod:\t{}{}.'
                   .format(character, prediction_method,
                           '\n\tModel:\t{}'.format(model) if model and is_ml(prediction_method) else ''))
        if COPY == prediction_method or is_parsimonious(prediction_method):
            states = get_states(prediction_method, model, character)
            character2settings[character] = [prediction_method, states]
        elif is_ml(prediction_method):
            params = column2parameters[character] if character in column2parameters else None
            rate_file = column2rates[character] if character in column2rates else None
            optimise_tau = tau is None or reoptimise
            if tau is None:
                tau = 0
            states = get_states(prediction_method, model, character)

            missing_data, observed_frequencies, state2index = calculate_observed_freqs(character, forest, states)

            logger = logging.getLogger('pastml')
            logger.debug('Observed frequencies for {}:{}{}.'
                         .format(character,
                                 ''.join('\n\tfrequency of {}:\t{:.6f}'
                                         .format(state, observed_frequencies[state2index[state]]) for state in states),
                                 '\n\tfraction of missing data:\t{:.6f}'
                                 .format(missing_data) if missing_data else '')
                         )

            model_instance = model2class[model](parameter_file=params, rate_matrix_file=rate_file, reoptimise=reoptimise,
                                                frequency_smoothing=frequency_smoothing, tau=tau,
                                                optimise_tau=optimise_tau, states=states, forest_stats=forest_stats,
                                                observed_frequencies=observed_frequencies)
            character2settings[character] = [prediction_method, model_instance]
        else:
            raise ValueError('Method {} is unknown, should be one of ML ({}), one of MP ({}) or {}'
                             .format(prediction_method, ', '.join(ML_METHODS), ', '.join(MP_METHODS), COPY))

    if threads < 1:
        threads = max(os.cpu_count(), 1)

    def _work(character):
        prediction_method, model_or_states = character2settings[character]
        if COPY == prediction_method:
            return {CHARACTER: character, STATES: model_or_states, METHOD: prediction_method}
        if is_ml(prediction_method):
            return ml_acr(forest=forest, character=character, prediction_method=prediction_method,
                          model=model_or_states,
                          force_joint=force_joint, observed_frequencies=observed_frequencies)
        if is_parsimonious(prediction_method):
            return parsimonious_acr(forest=forest, character=character, prediction_method=prediction_method,
                                    states=model_or_states,
                                    num_nodes=forest_stats.num_nodes, num_tips=forest_stats.num_tips)

    if threads > 1:
        with ThreadPool(processes=threads - 1) as pool:
            acr_results = \
                pool.map(func=_work, iterable=character2settings.keys())
    else:
        acr_results = [_work(character) for character in character2settings.keys()]

    acr_results = flatten_lists(acr_results)

    column2states = {acr_result[CHARACTER]: acr_result[STATES] for acr_result in acr_results}
    column2copy = {acr_result[CHARACTER]: acr_result[METHOD] == COPY for acr_result in acr_results}
    if resolve_polytomies and resolve_trees(column2states, forest):
        level = logger.level
        logger.setLevel(logging.ERROR)
        # we have selected states before, so now need to reset them
        for root in forest:
            for n in root.traverse():
                c2states = n2c2states[n]
                for c in columns:
                    if c in c2states:
                        n.add_feature(c, c2states[c])
                    # if it is a copy method we just need to keep the polytomy state
                    # as there is no way to calculate a state
                    elif not getattr(n, IS_POLYTOMY, False) or not column2copy[c]:
                        n.del_feature(c)

        forest_stats = ForestStats(forest)
        for acr_res in acr_results:
            character = acr_res[CHARACTER]
            method = acr_res[METHOD]
            if is_ml(method):
                if character not in character2settings:
                    character = character[:character.rfind('_{}').format(method)]
                character2settings[character][1].freeze()
                character2settings[character][1].forest_stats = forest_stats
        if threads > 1:
            with ThreadPool(processes=threads - 1) as pool:
                acr_results = \
                    pool.map(func=_work, iterable=character2settings.keys())
        else:
            acr_results = [_work(character) for character in character2settings.keys()]
        logger.setLevel(level)
        while unresolve_trees(column2states, forest):
            logger.setLevel(logging.ERROR)
            if threads > 1:
                with ThreadPool(processes=threads - 1) as pool:
                    acr_results = \
                        pool.map(func=_work, iterable=character2settings.keys())
            else:
                acr_results = [_work(character) for character in character2settings.keys()]
            logger.setLevel(level)
        logger.setLevel(level)
        acr_results = flatten_lists(acr_results)
    return acr_results


def calculate_observed_freqs(character, forest, states):
    n = len(states)
    missing_data = 0.
    state2index = dict(zip(states, range(n)))
    observed_frequencies = np.zeros(n, np.float64)
    for tree in forest:
        for _ in tree:
            state = getattr(_, character, set())
            if state:
                num_node_states = len(state)
                for _ in state:
                    observed_frequencies[state2index[_]] += 1. / num_node_states
            else:
                missing_data += 1
    total_count = observed_frequencies.sum() + missing_data
    observed_frequencies /= observed_frequencies.sum()
    missing_data /= total_count
    return missing_data, observed_frequencies, state2index


def flatten_lists(lists):
    result = []
    for _ in lists:
        if isinstance(_, list):
            result.extend(_)
        else:
            result.append(_)
    return result


def _quote(str_list):
    return ', '.join('"{}"'.format(_) for _ in str_list) if str_list is not None else ''


def pastml_pipeline(tree, data=None, data_sep='\t', id_index=0,
                    columns=None, prediction_method=MPPA, model=F81,
                    parameters=None, rate_matrix=None,
                    name_column=None, root_date=None, timeline_type=TIMELINE_SAMPLED,
                    tip_size_threshold=REASONABLE_NUMBER_OF_TIPS, colours=None,
                    out_data=None, html_compressed=None, html=None, html_mixed=None, work_dir=None,
                    verbose=False, forced_joint=False, upload_to_itol=False, itol_id=None, itol_project=None,
                    itol_tree_name=None, offline=False, threads=0, reoptimise=False, focus=None,
                    resolve_polytomies=False, smoothing=False, frequency_smoothing=False,
                    pajek=None, pajek_timing=VERTICAL, recursion_limit=0):
    """
    Applies PastML to the given tree(s) with the specified states and visualises the result (as html maps).

    :param tree: path to the input tree(s) in newick format (must be rooted).
    :type tree: str

    :param data: (optional) path to the annotation file in tab/csv format with the first row containing the column names.
        If not given, the annotations should be contained in the tree file itself.
    :type data: str
    :param data_sep: (optional, by default '\t') column separator for the annotation table.
        By default is set to tab, i.e. for tab-delimited file. Set it to ',' if your file is csv.
    :type data_sep: char
    :param id_index: (optional, by default is 0) index of the column in the annotation table
        that contains the tree tip names, indices start from zero.
    :type id_index: int

    :param columns: (optional) name(s) of the annotation table column(s) that contain character(s)
        to be analysed. If not specified all annotation table columns will be considered.
    :type columns: str or list(str)
    :param prediction_method: (optional, default is pastml.ml.MPPA) ancestral character reconstruction method(s),
        can be one of the max likelihood (ML) methods: pastml.ml.MPPA, pastml.ml.MAP, pastml.ml.JOINT,
        one of the max parsimony (MP) methods: pastml.parsimony.ACCTRAN, pastml.parsimony.DELTRAN,
        pastml.parsimony.DOWNPASS; or pastml.acr.COPY to keep the annotated character states as-is without inference.
        One can also specify one of the meta-methods: pastml.ml.ALL, pastml.ml.ML, pastml.parsimony.MP,
        that would perform ACR with multiple methods (all of them for pastml.ml.ALL,
        all the ML methods for pastml.ml.ML, or all the MP methods for pastml.parsimony.MP)
        and save/visualise the results as multiple characters suffixed with the corresponding method.
        When multiple ancestral characters are specified (with ``columns`` argument),
        the same method can be used for all of them (if only one method is specified),
        or different methods can be used (specified in the same order as ``columns``).
        If multiple methods are given, but not for all the characters,
        for the rest of them the default method (pastml.ml.MPPA) is chosen.'
    :type prediction_method: str or list(str)
    :param forced_joint: (optional, default is False) add JOINT state to the MPPA state selection
        even if it is not selected by Brier score.
    :type forced_joint: bool
    :param model: (optional, default is pastml.models.f81_like.F81) evolutionary model(s) for ML methods
        (ignored by MP methods).
        When multiple ancestral characters are specified (with ``columns`` argument),
        the same model can be used for all of them (if only one model is specified),
        or different models can be used (specified in the same order as ``columns``).
        If multiple models are given, but not for all the characters,
        for the rest of them the default model (pastml.models.f81_like.F81) is chosen.
    :type model: str or list(str)
    :param parameters: optional way to fix some of the ML-method parameters.
        Could be specified as
        (1a) a dict {column: {param: value}},
        where column corresponds to the character for which these parameters should be used,
        or (1b) in a form {column: path_to_param_file};
        or (2) as a list of paths to parameter files
        (in the same order as ``columns`` argument that specifies characters)
        possibly given only for the first few characters;
        or (3) as a path to parameter file (only for the first character).
        Each file should be tab-delimited, with two columns: the first one containing parameter names,
        and the second, named "value", containing parameter values.
        Parameters can include character state frequencies (parameter name should be the corresponding state,
        and parameter value - the float frequency value, between 0 and 1),
        tree branch scaling factor (parameter name pastml.ml.SCALING_FACTOR),
        and tree branch smoothing factor (parameter name pastml.ml.SMOOTHING_FACTOR).
    :type parameters: str or list(str) or dict
    :param rate_matrix: (only for pastml.models.rate_matrix.CUSTOM_RATES model) path to the file(s)
        specifying the rate matrix(ces).
        Could be specified as
        (1) a dict {column: path_to_file},
        where column corresponds to the character for which this rate matrix should be used,
        or (2) as a list of paths to rate matrix files
        (in the same order as ``columns`` argument that specifies characters)
        possibly given only for the first few characters;
        or (3) as a path to rate matrix file (only for the first character).
        The rate matrix file should specify character states in its first line, preceded by '# ' and separated by spaces.
        The following lines should contain a symmetric squared rate matrix with positive rates
        (and zeros on the diagonal), separated by spaces,
        in the same order at the character states specified in the first line.
        For example for four states, A, C, G, T and the rates A<->C 1, A<->G 4, A<->T 1, C<->G 1, C<->T 4, G<->T 1,
        the rate matrix file would look like:
        # A C G T
        0 1 4 1
        1 0 1 4
        4 1 0 1
        1 4 1 0
    :type rate_matrix: str or list(str) or dict
    :param reoptimise: (False by default) if set to True and the parameters are specified,
        they will be considered as an optimisation starting point instead, and optimised.
    :type reoptimise: bool
    :param smoothing: (optional, default is False) apply a smoothing factor (optimised) to branch lengths
        during likelihood calculation.
    :type smoothing: bool
    :param frequency_smoothing: (optional, default is False) apply a smoothing factor (optimised) to state frequencies
        (given as input parameters, see parameters argument) during likelihood calculation.
        If the selected model (model argument) does not allow for frequency optimisation, this option will be ignored.
        If reoptimise argument is also set to True, the frequencies will only be smoothed but not reoptimised.
    :type frequency_smoothing: bool
    :param name_column: (optional) name of the annotation table column to be used for node names
        in the compressed map visualisation
        (must be one of those specified in ``columns``, if ``columns`` are specified).
        If the annotation table contains only one column, it will be used by default.
    :type name_column: str
    :param root_date: (optional) date(s) of the root(s) (for dated tree(s) only),
        if specified, used to visualise a timeline based on dates (otherwise it is based on distances to root).
    :type root_date: str or pandas.datetime or float or list
    :param tip_size_threshold: (optional, by default is 15) recursively remove the tips
        of size less than threshold-th largest tip from the compressed map (set to 1e10 to keep all).
        The larger it is the less tips will be trimmed.
    :type tip_size_threshold: int
    :param focus: optional way to put a focus on certain character state values,
        so that the nodes in these states are displayed
        even if they do not pass the trimming threshold (tip_size_threshold argument).
        Should be in the form character:state.
    :type focus: str or list(str)
    :param timeline_type: (optional, by default is pastml.visualisation.cytoscape_manager.TIMELINE_SAMPLED)
        type of timeline visualisation: at each date/distance to root selected on the slider, either
        (pastml.visualisation.cytoscape_manager.TIMELINE_SAMPLED) all the lineages sampled after it are hidden; "
        or (pastml.visualisation.cytoscape_manager.TIMELINE_NODES) all the nodes with a
        more recent date/larger distance to root are hidden;
        or (pastml.visualisation.cytoscape_manager.TIMELINE_LTT) all the nodes whose branch started
        after this date/distance to root are hidden, and the external branches are cut
        to the specified date/distance to root if needed;
    :type timeline_type: str
    :param colours: optional way to specify the colours used for character state visualisation.
        Could be specified as
        (1a) a dict {column: {state: colour}},
        where column corresponds to the character for which these parameters should be used,
        or (1b) in a form {column: path_to_colour_file};
        or (2) as a list of paths to colour files
        (in the same order as ``columns`` argument that specifies characters)
        possibly given only for the first few characters;
        or (3) as a path to colour file (only for the first character).
        Each file should be tab-delimited, with two columns: the first one containing character states,
        and the second, named "colour", containing colours, in HEX format (e.g. #a6cee3).
    :type colours: str or list(str) or dict
    :param resolve_polytomies: (default False) when True, the polytomies with a state change
        (i.e. a parent node, P, in state A has more than 2 children, including m > 1 children, C_1, ..., C_m, in state B)
        are resolved by grouping together same-state (different from the parent state) nodes
        (i.e. a new internal node N in state B is created and becomes the child of P and the parent of C_1, ..., C_m).
    :type resolve_polytomies: bool

    :param out_data: path to the output annotation file with the reconstructed ancestral character states.
    :type out_data: str
    :param html_compressed: path to the output compressed visualisation file (html).
    :type html_compressed: str
    :param pajek: path to the output compressed visualisation file (Pajek NET Format).
        Produced only if html_compressed is specified.
    :type pajek: str
    :param pajek_timing: the type of the compressed visualisation to be saved in Pajek NET Format (if pajek is specified).
        Can be either 'VERTICAL' (default, after the nodes underwent vertical compression),
        'HORIZONTAL' (after the nodes underwent vertical and horizontal compression)
        or 'TRIM' (after the nodes underwent vertical and horizontal compression and minor node trimming).
    :type pajek_timing: str
    :param html: (optional) path to the output tree visualisation file (html).
    :type html: str
    :param html_mixed: (optional) path to the output mostly compressed map visualisation file (html),
        where the nodes in states specified with the focus argument are uncompressed.
    :type html_mixed: str
    :param work_dir: (optional) path to the folder where pastml parameter, named tree
        and marginal probability (for marginal ML methods (pastml.ml.MPPA, pastml.ml.MAP) only) files are to be stored.
        Default is <path_to_input_file>/<input_file_name>_pastml. If the folder does not exist, it will be created.
    :type work_dir: str
    :param offline: (optional, default is False) By default (offline=False) PastML assumes
        that there is an internet connection available,
        which permits it to fetch CSS and JS scripts needed for visualisation online.
        With offline=True, PastML will store all the needed CSS/JS scripts in the folder specified by work_dir,
        so that internet connection is not needed
        (but you must not move the output html files to any location other that the one specified by html/html_compressed.
    :type offline: bool

    :param verbose: (optional, default is False) print information on the progress of the analysis.
    :type verbose: bool

    :param threads: (optional, default is 0, which stands for automatic) number of threads PastML can use for parallesation.
        By default, detected automatically based on the system. Note that PastML will at most use as many threads
        as the number of characters (-c option) being analysed plus one.
    :type threads: int

    :parameter recursion_limit: Set the recursion limit in python.
        If your tree is very large and pastml produces an overflow error,
        you could try to set the recursion limit to a large number (e.g., 10000).
        By default the standard recursion limit is used and should be sufficient
        for trees of up to several thousand tips.
    :type recursion_limit: int

    :param upload_to_itol: (optional, default is False) whether iTOL annotations
        for the reconstructed characters associated with the named tree (i.e. the one found in work_dir) should be created.
        If additionally itol_id and itol_project are specified,
        the annotated tree will be automatically uploaded to iTOL (https://itol.embl.de/).
    :type upload_to_itol: bool
    :param itol_id: (optional) iTOL user batch upload ID that enables uploading to your iTOL account
        (see https://itol.embl.de/help.cgi#batch).
    :type itol_id: str
    :param itol_project: (optional) iTOL project the annotated tree should be uploaded to
        (must exist, and itol_id must be specified). If not specified, the tree will not be associated to any project.
    :type itol_project: str
    :param itol_tree_name: (optional) name for the tree uploaded to iTOL.
    :type itol_tree_name: str

    :return: void
    """
    logger = _set_up_pastml_logger(verbose)

    if recursion_limit and recursion_limit > 0:
        sys.setrecursionlimit(recursion_limit)

    copy_only = COPY == prediction_method or (isinstance(prediction_method, list)
                                              and all(COPY == _ for _ in prediction_method))

    roots, columns, column2states, name_column, age_label, parameters, rates = \
        _validate_input(tree, columns, name_column if html_compressed or html_mixed else None, data, data_sep, id_index,
                        root_date if html_compressed or html or html_mixed or upload_to_itol else None,
                        copy_only=copy_only, parameters=parameters, rates=rate_matrix)

    if not work_dir:
        work_dir = get_pastml_work_dir(tree)
    os.makedirs(work_dir, exist_ok=True)

    if threads < 1:
        threads = max(os.cpu_count(), 1)

    acr_results = acr(forest=roots, columns=columns, column2states=column2states,
                      prediction_method=prediction_method, model=model, column2parameters=parameters,
                      column2rates=rates,
                      force_joint=forced_joint, threads=threads, reoptimise=reoptimise, tau=None if smoothing else 0,
                      resolve_polytomies=resolve_polytomies, frequency_smoothing=frequency_smoothing)

    column2states = {acr_result[CHARACTER]: acr_result[STATES] for acr_result in acr_results}

    if not out_data:
        out_data = os.path.join(work_dir, get_combined_ancestral_state_file())

    state_df = _serialize_predicted_states(sorted(column2states.keys()), out_data, roots,
                                           dates_are_dates=age_label == DATE_LABEL)

    # a meta-method would have added a suffix to the name feature
    if html_compressed and name_column and name_column not in column2states:
        ml_name_column = get_personalized_feature_name(name_column, get_default_ml_method())
        name_column = ml_name_column if ml_name_column in column2states \
            else get_personalized_feature_name(name_column, get_default_mp_method())

    itol_result = None
    new_tree = os.path.join(work_dir, get_named_tree_file(tree))
    features = [DATE, DATE_CI] + list(column2states.keys())
    clear_extra_features(roots, features)
    nwks = '\n'.join([root.write(format_root_node=True, format=3, features=features) for root in roots])
    with open(new_tree, 'w+') as f:
        f.write(nwks)
    try:
        nexus = new_tree.replace('.nwk', '.nexus')
        if '.nexus' not in nexus:
            nexus = '{}.nexus'.format(nexus)
        write(NewickIO.parse(StringIO(nwks)), nexus, 'nexus')
        with open(nexus, 'r') as f:
            nexus_str = f.read().replace('&&NHX:', '&')
            for feature in features:
                nexus_str = nexus_str.replace(':{}='.format(feature), ',{}='.format(feature))
        with open(nexus, 'w') as f:
            f.write(nexus_str)
    except Exception as e:
        logger.error(
            'Did not manage to save the annotated tree in nexus format due to the following error: {}'.format(e))
        pass

    if upload_to_itol or html or html_compressed:
        if colours:
            if isinstance(colours, str):
                colours = [colours]
            if isinstance(colours, list):
                colours = dict(zip(columns, colours))
            elif isinstance(colours, dict):
                colours = {col_name2cat(col): cls for (col, cls) in colours.items()}
            else:
                raise ValueError('Colours should be either a list or a dict, got {}.'.format(type(colours)))
        else:
            colours = {}

    if threads > 1:
        pool = ThreadPool(processes=threads - 1)
        async_result = pool.map_async(func=_serialize_acr, iterable=((acr_res, work_dir) for acr_res in acr_results))
        if upload_to_itol:
            if DATE_LABEL == age_label:
                try:
                    dates = state_df[DATE].apply(lambda _: numeric2datetime(_).strftime("%d %b %Y"))
                    state_df[DATE] = dates
                except:
                    pass
            itol_result = pool.apply_async(func=generate_itol_annotations,
                                           args=(column2states, work_dir, acr_results, state_df, age_label,
                                                 new_tree, itol_id, itol_project, itol_tree_name, colours))
    else:
        for acr_res in acr_results:
            _serialize_acr((acr_res, work_dir))
        if upload_to_itol:
            if DATE_LABEL == age_label:
                try:
                    dates = state_df[DATE].apply(lambda _: numeric2datetime(_).strftime("%d %b %Y"))
                    state_df[DATE] = dates
                except:
                    pass
            generate_itol_annotations(column2states, work_dir, acr_results, state_df, age_label,
                                      new_tree, itol_id, itol_project, itol_tree_name, colours)

    if html or html_compressed or html_mixed:
        logger.debug('\n=============VISUALISATION=====================')

        if (html_compressed or html_mixed) and focus:
            def parse_col_val(cv):
                cv = str(cv).strip()
                colon_pos = cv.find(':')
                if colon_pos == -1:
                    if len(column2states) == 1 and cv in next(iter(column2states.values())):
                        return next(iter(column2states.keys())), cv
                    else:
                        raise ValueError('Focus values should be in a form character:state, got {} instead.'.format(cv))
                col, state = col_name2cat(cv[:colon_pos]), cv[colon_pos + 1:]
                if col not in column2states:
                    ml_col = get_personalized_feature_name(col, get_default_ml_method())
                    if ml_col in column2states:
                        col = ml_col
                    else:
                        mp_col = get_personalized_feature_name(col, get_default_mp_method())
                        if mp_col in column2states:
                            col = mp_col
                        else:
                            raise ValueError('Character {} specified for focus values is not found in metadata.'.format(
                                cv[:colon_pos]))
                if state not in column2states[col]:
                    raise ValueError(
                        'Character {} state {} not found among possible states in metadata.'.format(cv[:colon_pos],
                                                                                                    state))
                return col, state

            if isinstance(focus, str):
                focus = list(focus)
            if not isinstance(focus, list):
                raise ValueError(
                    'Focus argument should be either a string or a list of strings, got {} instead.'.format(
                        type(focus)))
            focus_cv = [parse_col_val(_) for _ in focus]
            focus = defaultdict(set)
            for c, v in focus_cv:
                focus[c].add(v)

        visualize(roots, column2states=column2states, html=html, html_compressed=html_compressed, html_mixed=html_mixed,
                  name_column=name_column, tip_size_threshold=tip_size_threshold, date_label=age_label,
                  timeline_type=timeline_type, work_dir=work_dir, local_css_js=offline, column2colours=colours,
                  focus=focus, pajek=pajek, pajek_timing=pajek_timing)

    if threads > 1:
        async_result.wait()
        if upload_to_itol:
            itol_result.wait()
        pool.close()


def _validate_input(tree_nwk, columns=None, name_column=None, data=None, data_sep='\t', id_index=0,
                    root_dates=None, copy_only=False, parameters=None, rates=None):
    logger = logging.getLogger('pastml')
    logger.debug('\n=============INPUT DATA VALIDATION=============')

    if not columns and data is None:
        raise ValueError("If you don't provide the metadata file, "
                         "you need to provide an annotated tree and specify the columns argument, "
                         "which will be used to look for character annotations in your input tree.")

    if columns and isinstance(columns, str):
        columns = [columns]

    roots = read_forest(tree_nwk, columns=columns if data is None else None)

    column2annotated = Counter()
    column2states = defaultdict(set)

    if data:
        df = pd.read_csv(data, sep=data_sep, index_col=id_index, header=0, dtype=str)
        df.index = df.index.map(str)
        logger.debug('Read the annotation file {}.'.format(data))
        if columns:
            unknown_columns = set(columns) - set(df.columns)
            if unknown_columns:
                raise ValueError('{} of the specified columns ({}) {} not found among the annotation columns: {}.'
                                 .format('One' if len(unknown_columns) == 1 else 'Some',
                                         _quote(unknown_columns),
                                         'is' if len(unknown_columns) == 1 else 'are',
                                         _quote(df.columns)))
            df = df[columns]
        df.columns = [col_name2cat(column) for column in df.columns]
        if name_column:
            name_column = col_name2cat(name_column)
        columns = df.columns

        node_names = set.union(*[{n.name for n in root.traverse() if n.name} for root in roots])
        df_index_names = set(df.index)
        common_ids = node_names & df_index_names

        # strip quotes if needed
        if not common_ids:
            node_names = {_.strip("'").strip('"') for _ in node_names}
            common_ids = node_names & df_index_names
            if common_ids:
                for root in roots:
                    for n in root.traverse():
                        n.name = n.name.strip("'").strip('"')

        filtered_df = df.loc[list(common_ids), :]
        if not filtered_df.shape[0]:
            tip_name_representatives = []
            for _ in roots[0].iter_leaves():
                if len(tip_name_representatives) < 3:
                    tip_name_representatives.append(_.name)
                else:
                    break
            raise ValueError(
                'Your tree tip names (e.g. {}) do not correspond to annotation id column values (e.g. {}). '
                'Check your annotation file.'
                    .format(', '.join(tip_name_representatives),
                            ', '.join(list(df_index_names)[: min(len(df_index_names), 3)])))
        logger.debug('Checked that (at least some of) tip names correspond to annotation file index.')
        preannotate_forest(roots, df=df)
        for c in df.columns:
            column2states[c] |= {_ for _ in df[c].unique() if pd.notnull(_) and _ != ''}

    num_tips = 0

    column2annotated_states = defaultdict(set)
    for root in roots:
        for n in root.traverse():
            for c in columns:
                vs = getattr(n, c, set())
                column2states[c] |= vs
                column2annotated_states[c] |= vs
                if vs:
                    column2annotated[c] += 1
            if n.is_leaf():
                num_tips += 1

    if column2annotated:
        c, num_annotated = min(column2annotated.items(), key=lambda _: _[1])
    else:
        c, num_annotated = columns[0], 0
    percentage_unknown = (num_tips - num_annotated) / num_tips
    if percentage_unknown >= (.9 if not copy_only else 1):
        raise ValueError('{:.1f}% of tip annotations for character "{}" are unknown, '
                         'not enough data to infer ancestral states. '
                         '{}'
                         .format(percentage_unknown * 100, c,
                                 'Check your annotation file and if its ids correspond to the tree tip/node names.'
                                 if data
                                 else 'You tree file should contain character state annotations, '
                                      'otherwise consider specifying a metadata file.'))

    c, states = min(column2annotated_states.items(), key=lambda _: len(_[1]))
    if len(states) > num_tips * .75 and not copy_only:
        raise ValueError('Character "{}" has {} unique states annotated in this tree: {}, '
                         'which is too much to infer on a {} with only {} tips. '
                         'Make sure the character you are analysing is discrete, and if yes use a larger tree.'
                         .format(c, len(states), states, 'tree' if len(roots) == 1 else 'forest', num_tips))

    if name_column and name_column not in columns:
        raise ValueError('The name column ("{}") should be one of those specified as columns ({}).'
                         .format(name_column, _quote(columns)))
    elif len(columns) == 1:
        name_column = columns[0]

    # Process root dates
    if root_dates is not None:
        root_dates = [parse_date(d) for d in (root_dates if isinstance(root_dates, list) else [root_dates])]
        if 1 < len(root_dates) < len(roots):
            raise ValueError('{} trees are given, but only {} root dates.'.format(len(roots), len(root_dates)))
        elif 1 == len(root_dates):
            root_dates *= len(roots)
    age_label = DIST_TO_ROOT_LABEL \
        if (root_dates is None and not next((True for root in roots if getattr(root, DATE, None) is not None), False)) \
        else DATE_LABEL
    annotate_dates(roots, root_dates=root_dates)
    logger.debug('Finished input validation.')

    column2states = {c: np.array(sorted(states)) for c, states in column2states.items()}

    if parameters:
        if isinstance(parameters, str):
            parameters = [parameters]
        if isinstance(parameters, list):
            parameters = dict(zip(columns, parameters))
        elif isinstance(parameters, dict):
            parameters = {col_name2cat(col): params for (col, params) in parameters.items()}
        else:
            raise ValueError('Parameters should be either a list or a dict, got {}.'.format(type(parameters)))
    else:
        parameters = {}

    if rates:
        if isinstance(rates, str):
            rates = [rates]
        if isinstance(rates, list):
            rates = dict(zip(columns, rates))
        elif isinstance(rates, dict):
            rates = {col_name2cat(col): rs for (col, rs) in rates.items()}
        else:
            raise ValueError('Rate matrices should be either a list or a dict, got {}.'.format(type(rates)))
    else:
        rates = {}

    for i, tree in enumerate(roots):
        name_tree(tree, suffix='' if len(roots) == 1 else '_{}'.format(i))

    return roots, columns, column2states, name_column, age_label, parameters, rates


def _serialize_predicted_states(columns, out_data, roots, dates_are_dates=True):
    ids, data = [], []
    # Not using DataFrames to speed up document writing
    with open(out_data, 'w+') as f:
        f.write('node\t{}\n'.format('\t'.join(columns)))
        for root in roots:
            for node in root.traverse():
                vs = [node.dist, get_formatted_date(node, dates_are_dates)]
                column2values = {}
                for column in columns:
                    value = getattr(node, column, set())
                    vs.append(value)
                    if value:
                        column2values[column] = sorted(value, reverse=True)
                data.append(vs)
                ids.append(node.name)
                while column2values:
                    f.write('{}'.format(node.name))
                    for column in columns:
                        if column in column2values:
                            values = column2values[column]
                            value = values.pop()
                            if not values:
                                del column2values[column]
                        else:
                            value = ''
                        f.write('\t{}'.format(value))
                    f.write('\n')
    logging.getLogger('pastml').debug('Serialized reconstructed states to {}.'.format(out_data))
    return pd.DataFrame(index=ids, data=data, columns=['dist', DATE] + columns)


def main():
    """
    Entry point, calling :py:func:`pastml.acr.pastml_pipeline` with command-line arguments.

    :return: void
    """
    import argparse

    parser = argparse.ArgumentParser(description="Ancestral character reconstruction and visualisation "
                                                 "for rooted phylogenetic trees.", prog='pastml')

    tree_group = parser.add_argument_group('tree-related arguments')
    tree_group.add_argument('-t', '--tree', help="input tree(s) in newick format (must be rooted).",
                            type=str, required=True)

    annotation_group = parser.add_argument_group('annotation-file-related arguments')
    annotation_group.add_argument('-d', '--data', required=False, type=str, default=None,
                                  help="annotation file in tab/csv format with the first row "
                                       "containing the column names. "
                                       "If not given, the annotations should be contained in the tree file itself.")
    annotation_group.add_argument('-s', '--data_sep', required=False, type=str, default='\t',
                                  help="column separator for the annotation table. "
                                       "By default is set to tab, i.e. for a tab-delimited file. "
                                       "Set it to ',' if your file is csv.")
    annotation_group.add_argument('-i', '--id_index', required=False, type=int, default=0,
                                  help="index of the annotation table column containing tree tip names, "
                                       "indices start from zero (by default is set to 0).")

    acr_group = parser.add_argument_group('ancestral-character-reconstruction-related arguments')
    acr_group.add_argument('-c', '--columns', nargs='*',
                           help="names of the annotation table columns that contain characters "
                                "to be analysed. "
                                "If not specified, all columns are considered.",
                           type=str)
    acr_group.add_argument('--prediction_method',
                           choices=[MPPA, MAP, JOINT, DOWNPASS, ACCTRAN, DELTRAN, COPY, ALL, ML, MP],
                           type=str, nargs='*', default=MPPA,
                           help='ancestral character reconstruction (ACR) method, '
                                'can be one of the max likelihood (ML) methods: {ml}, '
                                'one of the max parsimony (MP) methods: {mp}; '
                                'or {copy} to keep the annotated character states as-is without inference. '
                                'One can also specify one of the meta-methods {meta} that would perform ACR '
                                'with multiple methods (all of them for {meta_all}, '
                                'all the ML methods for {meta_ml}, or all the MP methods for {meta_mp}) '
                                'and save/visualise the results as multiple characters '
                                'suffixed with the corresponding method.'
                                'When multiple ancestral characters are specified (see -c, --columns), '
                                'the same method can be used for all of them (if only one method is specified), '
                                'or different methods can be used (specified in the same order as -c, --columns). '
                                'If multiple methods are given, but not for all the characters, '
                                'for the rest of them the default method ({default}) is chosen.'
                           .format(ml=', '.join(ML_METHODS), mp=', '.join(MP_METHODS), copy=COPY, default=MPPA,
                                   meta=', '.join(META_ML_METHODS | {MP}), meta_ml=ML, meta_mp=MP, meta_all=ALL))
    acr_group.add_argument('--forced_joint', action='store_true',
                           help='add {joint} state to the {mppa} state selection '
                                'even if it is not selected by Brier score.'.format(joint=JOINT, mppa=MPPA))
    acr_group.add_argument('-m', '--model', default=F81,
                           choices=[JC, F81, EFT, HKY, JTT, CUSTOM_RATES],
                           type=str, nargs='*',
                           help='evolutionary model for ML methods (ignored by MP methods). '
                                'When multiple ancestral characters are specified (see -c, --columns), '
                                'the same model can be used for all of them (if only one model is specified), '
                                'or different models can be used (specified in the same order as -c, --columns). '
                                'If multiple models are given, but not for all the characters, '
                                'for the rest of them the default model ({}) is chosen.'.format(F81))
    acr_group.add_argument('--parameters', type=str, nargs='*',
                           help='optional way to fix some of the ML-method parameters '
                                'by specifying files that contain them. '
                                'Should be in the same order '
                                'as the ancestral characters (see -c, --columns) '
                                'for which the reconstruction is to be preformed. '
                                'Could be given only for the first few characters. '
                                'Each file should be tab-delimited, with two columns: '
                                'the first one containing parameter names, '
                                'and the second, named "value", containing parameter values. '
                                'Parameters can include character state frequencies '
                                '(parameter name should be the corresponding state, '
                                'and parameter value - the float frequency value, between 0 and 1),'
                                'tree branch scaling factor (parameter name {}),'.format(SCALING_FACTOR) +
                                'and tree branch smoothing factor (parameter name {}),'.format(SMOOTHING_FACTOR))
    acr_group.add_argument('--rate_matrix', type=str, nargs='*',
                           help='(only for {} model) path to the file(s) containing the rate matrix(ces). '
                                'Should be in the same order '
                                'as the ancestral characters (see -c, --columns) '
                                'for which the reconstruction is to be preformed. '
                                'Could be given only for the first few characters. '
                                'The rate matrix file should specify character states in its first line, '
                                'preceded by #  and separated by spaces. '
                                'The following lines should contain a symmetric squared rate matrix with positive rates'
                                '(and zeros on the diagonal), separated by spaces, '
                                'in the same order at the character states specified in the first line.'
                                'For example, for four states, A, C, G, T '
                                'and the rates A<->C 1, A<->G 4, A<->T 1, C<->G 1, C<->T 4, G<->T 1,'
                                'the rate matrix file would look like:\n'
                                '# A C G T\n'
                                '0 1 4 1\n'
                                '1 0 1 4\n'
                                '4 1 0 1\n'
                                '1 4 1 0'.format(CUSTOM_RATES))
    acr_group.add_argument('--reoptimise', action='store_true',
                           help='if the parameters are specified, they will be considered as an optimisation '
                                'starting point instead and optimised.')
    acr_group.add_argument('--smoothing', action='store_true',
                           help='Apply a smoothing factor (optimised) to branch lengths during likelihood calculation.')
    acr_group.add_argument('--frequency_smoothing', action='store_true',
                           help='Apply a smoothing factor (optimised) to state frequencies '
                                '(given as input parameters, see --parameters) '
                                'during likelihood calculation. '
                                'If the selected model (--model) does not allow for frequency optimisation,'
                                ' this option will be ignored. '
                                'If --reoptimise is also specified, '
                                'the frequencies will only be smoothed but not reoptimised. ')

    vis_group = parser.add_argument_group('visualisation-related arguments')
    vis_group.add_argument('-n', '--name_column', type=str, default=None,
                           help="name of the character to be used for node names "
                                "in the compressed map visualisation "
                                "(must be one of those specified via -c, --columns). "
                                "If the annotation table contains only one column it will be used by default.")
    vis_group.add_argument('--root_date', required=False, default=None,
                           help="date(s) of the root(s) (for dated tree(s) only), "
                                "if specified, used to visualise a timeline based on dates "
                                "(otherwise it is based on distances to root).",
                           type=str, nargs='*')
    vis_group.add_argument('--tip_size_threshold', type=int, default=REASONABLE_NUMBER_OF_TIPS,
                           help="recursively remove the tips of size less than threshold-th largest tip"
                                "from the compressed map (set to 1e10 to keep all tips). "
                                "The larger it is the less tips will be trimmed.")
    vis_group.add_argument('--timeline_type', type=str, default=TIMELINE_SAMPLED,
                           help="type of timeline visualisation: at each date/distance to root selected on the slider "
                                "either ({sampled}) - all the lineages sampled after it are hidden; "
                                "or ({nodes}) - all the nodes with a more recent date/larger distance to root are hidden; "
                                "or ({ltt}) - all the nodes whose branch started after this date/distance to root "
                                "are hidden, and the external branches are cut to the specified date/distance to root "
                                "if needed;".format(sampled=TIMELINE_SAMPLED, ltt=TIMELINE_LTT, nodes=TIMELINE_NODES),
                           choices=[TIMELINE_SAMPLED, TIMELINE_NODES, TIMELINE_LTT])
    vis_group.add_argument('--offline', action='store_true',
                           help="By default (without --offline option) PastML assumes "
                                "that there is an internet connection available, "
                                "which permits it to fetch CSS and JS scripts needed for visualisation online."
                                "With --offline option turned on, PastML will store all the needed CSS/JS scripts "
                                "in the folder specified by --work_dir, so that internet connection is not needed "
                                "(but you must not move the output html files to any location "
                                "other that the one specified by --html/--html_compressed).")
    vis_group.add_argument('--colours', type=str, nargs='*',
                           help='optional way to specify the colours used for character state visualisation. '
                                'Should be in the same order '
                                'as the ancestral characters (see -c, --columns) '
                                'for which the reconstruction is to be preformed. '
                                'Could be given only for the first few characters. '
                                'Each file should be tab-delimited, with two columns: '
                                'the first one containing character states, '
                                'and the second, named "colour", containing colours, in HEX format (e.g. #a6cee3).')
    vis_group.add_argument('--focus', type=str, nargs='*',
                           help='optional way to put a focus on certain character state values, '
                                'so that the nodes in these states are displayed '
                                'even if they do not pass the trimming threshold (--tip_size_threshold). '
                                'Should be in the form character:state.')
    vis_group.add_argument('--resolve_polytomies', action='store_true',
                           help='When specified, the polytomies with a state change '
                                '(i.e. a parent node, P, in state A has more than 2 children, '
                                'including m > 1 children, C_1, ..., C_m, in state B) are resolved '
                                'by grouping together same-state (different from the parent state) nodes '
                                '(i.e. a new internal node N in state B is created and becomes the child of P '
                                'and the parent of C_1, ..., C_m).')

    out_group = parser.add_argument_group('output-related arguments')
    out_group.add_argument('-o', '--out_data', required=False, type=str,
                           help="path to the output annotation file with the reconstructed ancestral character states.")
    out_group.add_argument('--work_dir', required=False, default=None, type=str,
                           help="path to the folder where pastml parameter, named tree "
                                "and marginal probability (for marginal ML methods ({}) only) files are to be stored. "
                                "Default is <path_to_input_file>/<input_file_name>_pastml. "
                                "If the folder does not exist, it will be created."
                           .format(', '.join(MARGINAL_ML_METHODS)))
    out_group.add_argument('--html_compressed', required=False, default=None, type=str,
                           help="path to the output compressed map visualisation file (html).")
    out_group.add_argument('--pajek', required=False, default=None, type=str,
                           help="path to the output vertically compressed visualisation file (Pajek NET Format). "
                                "Prooduced only if --html_compressed is specified.")
    out_group.add_argument('--pajek_timing', required=False, default=VERTICAL, choices=(VERTICAL, HORIZONTAL, TRIM),
                           type=str,
                           help="the type of the compressed visualisation to be saved in Pajek NET Format "
                                "(if --pajek is specified). "
                                "Can be either {} (default, after the nodes underwent vertical compression), "
                                "{} (after the nodes underwent vertical and horizontal compression) "
                                "or {} (after the nodes underwent vertical and horizontal compression"
                                " and minor node trimming)".format(VERTICAL, HORIZONTAL, TRIM))
    out_group.add_argument('--html', required=False, default=None, type=str,
                           help="path to the output full tree visualisation file (html).")
    out_group.add_argument('--html_mixed', required=False, default=None, type=str,
                           help="path to the output mostly compressed map visualisation file (html), "
                                "where the nodes in states specified with --focus are uncompressed.")
    out_group.add_argument('-v', '--verbose', action='store_true',
                           help="print information on the progress of the analysis (to console)")

    parser.add_argument('--version', action='version',
                        version='%(prog)s {version}'.format(version=PASTML_VERSION))

    parser.add_argument('--threads', required=False, default=0, type=int,
                        help="Number of threads PastML can use for parallesation. "
                             "By default detected automatically based on the system. "
                             "Note that PastML will at most use as many threads "
                             "as the number of characters (-c option) being analysed plus one.")
    parser.add_argument('--recursion_limit', required=False, default=0, type=int,
                        help='Set the recursion limit in python. '
                             'If your tree is very large and pastml produces an overflow error, '
                             'you could try to set the recursion limit to a large number (e.g., 10000). '
                             'By default the standard recursion limit is used and should be sufficient '
                             'for trees of up to several thousand tips.')

    itol_group = parser.add_argument_group('iTOL-related arguments')
    itol_group.add_argument('--upload_to_itol', action='store_true',
                            help="create iTOL annotations for the reconstructed characters "
                                 "associated with the named tree (i.e. the one found in --work_dir). "
                                 "If additionally --itol_id and --itol_project are specified, "
                                 "the annotated tree will be automatically uploaded to iTOL (https://itol.embl.de/).")
    itol_group.add_argument('--itol_id', required=False, default=None, type=str,
                            help="iTOL user batch upload ID that enables uploading to your iTOL account "
                                 "(see https://itol.embl.de/help.cgi#batch).")
    itol_group.add_argument('--itol_project', required=False, default="Sample project", type=str,
                            help="iTOL project the annotated tree should be associated with "
                                 "(must exist, and --itol_id must be specified). By default set to 'Sample project'.")
    itol_group.add_argument('--itol_tree_name', required=False, default=None, type=str,
                            help="name for the tree uploaded to iTOL.")

    params = parser.parse_args()

    pastml_pipeline(**vars(params))


if '__main__' == __name__:
    main()



import logging
import os
from collections import defaultdict
from glob import glob
from queue import Queue
from shutil import copyfile

import numpy as np
from jinja2 import Environment, PackageLoader

from pastml import numeric2datetime
from pastml.file import get_pastml_colour_file
from pastml.tree import DATE, IS_POLYTOMY, copy_forest
from pastml.visualisation import get_formatted_date
from pastml.visualisation.colour_generator import get_enough_colours, WHITE, parse_colours
from pastml.visualisation.tree_compressor import NUM_TIPS_INSIDE, TIPS_INSIDE, TIPS_BELOW, \
    REASONABLE_NUMBER_OF_TIPS, compress_tree, INTERNAL_NODES_INSIDE, ROOTS, IS_TIP, ROOT_DATES, IN_FOCUS, AROUND_FOCUS, \
    UP_FOCUS, save_to_pajek, VERTICAL

JS_LIST = ["https://pastml.pasteur.fr/static/js/jquery.min.js",
           "https://pastml.pasteur.fr/static/js/jquery.qtip.min.js",
           "https://pastml.pasteur.fr/static/js/cytoscape.min.js",
           "https://pastml.pasteur.fr/static/js/cytoscape-qtip.js",
           "https://pastml.pasteur.fr/static/js/cytoscape-svg.js",
           "https://pastml.pasteur.fr/static/js/layout-base.min.js",
           "https://pastml.pasteur.fr/static/js/cose-base.min.js",
           "https://pastml.pasteur.fr/static/js/cytoscape-cose-bilkent.min.js"]
CSS_LIST = ["https://pastml.pasteur.fr/static/css/jquery.qtip.min.css",
            "https://pastml.pasteur.fr/static/css/bootstrap.min.css"]

MAX_TIPS_FOR_FULL_TREE_VISUALISATION = 5000

TIMELINE_SAMPLED = 'SAMPLED'
TIMELINE_NODES = 'NODES'
TIMELINE_LTT = 'LTT'

TIP_LIMIT = 1000

MIN_EDGE_SIZE = 50
MIN_FONT_SIZE = 80
MIN_NODE_SIZE = 200

UNRESOLVED = 'unresolved'
TIP = 'tip'

TOOLTIP = 'tooltip'
COLOUR = 'colour'

DATA = 'data'
ID = 'id'
EDGES = 'edges'
NODES = 'nodes'
ELEMENTS = 'elements'

NODE_SIZE = 'node_size'
NODE_NAME = 'node_name'
BRANCH_NAME = 'branch_name'
EDGE_SIZE = 'edge_size'
EDGE_NAME = 'edge_name'
FONT_SIZE = 'node_fontsize'

MILESTONE = 'mile'

DATE_LABEL = 'date'

DIST_TO_ROOT_LABEL = 'dist. to root'


def get_fake_node(n_id, x, y):
    attributes = {ID: n_id, 'fake': 1}
    return _get_node(attributes, position=(x, y))


def get_node(n, n_id, tooltip='', clazz=None, x=0, y=0):
    features = {feature: getattr(n, feature) for feature in n.features if feature in [MILESTONE, UNRESOLVED, 'x', 'y']
                or feature.startswith('node_')}
    features[ID] = n_id
    if n.is_leaf():
        features[TIP] = 1
    features[TOOLTIP] = tooltip
    return _get_node(features, clazz=_clazz_list2css_class(clazz), position=(x, y) if x is not None else None)


def get_edge(source_name, target_name, **kwargs):
    return _get_edge(source=source_name, target=target_name, **kwargs)


def get_scaling_function(y_m, y_M, x_m, x_M):
    """
    Returns a linear function y = k x + b, where y \in [m, M]
    :param y_m:
    :param y_M:
    :param x_m:
    :param x_M:
    :return:
    """
    if x_M <= x_m:
        return lambda _: y_m
    k = (y_M - y_m) / (x_M - x_m)
    b = y_m - k * x_m
    return lambda _: int(k * _ + b)


def set_cyto_features_compressed(n, size_scaling, e_size_scaling, font_scaling, transform_size, transform_e_size, state,
                                 root_names, root_dates, suffix='', is_mixed=False):
    tips_inside, tips_below, internal_nodes_inside, roots = \
        getattr(n, TIPS_INSIDE, []), getattr(n, TIPS_BELOW, []), \
        getattr(n, INTERNAL_NODES_INSIDE, []), getattr(n, ROOTS, [])

    def get_min_max_str(values, default_value=0):
        min_v, max_v = (min(len(_) for _ in values), max(len(_) for _ in values)) \
            if values else (default_value, default_value)
        return ' {}'.format('{}-{}'.format(min_v, max_v) if min_v != max_v else min_v), min_v, max_v

    tips_below_str, _, max_n_tips_below = get_min_max_str(tips_below)
    tips_inside_str, _, max_n_tips = get_min_max_str(tips_inside)
    internal_ns_inside_str, _, _ = get_min_max_str(internal_nodes_inside)
    n.add_feature('{}{}'.format(NODE_NAME, suffix),
                  '{}{}'.format(state, tips_inside_str)
                  if not is_mixed or (not getattr(n, IN_FOCUS, False) and not getattr(n, UP_FOCUS, False)) else
                  '{}{}{}'.format(state, ':' if state else '', root_names[0]))
    size_factor = 2 if getattr(n, UNRESOLVED, False) else 1
    n.add_feature('{}{}'.format(NODE_SIZE, suffix),
                  (size_scaling(transform_size(max_n_tips)) if max_n_tips else int(MIN_NODE_SIZE / 1.5))
                  * size_factor)
    n.add_feature('{}{}'.format(FONT_SIZE, suffix),
                  font_scaling(transform_size(max_n_tips)) if max_n_tips else MIN_FONT_SIZE)

    n.add_feature('node_{}{}'.format(TIPS_INSIDE, suffix), tips_inside_str)
    n.add_feature('node_{}{}'.format(INTERNAL_NODES_INSIDE, suffix), internal_ns_inside_str)
    n.add_feature('node_{}{}'.format(TIPS_BELOW, suffix), tips_below_str)
    root_name2date = dict(zip(root_names, root_dates))
    root_names = sorted(root_names)
    n.add_feature('node_{}{}'.format(ROOTS, suffix), ', '.join(root_names))
    n.add_feature('node_{}{}'.format(ROOT_DATES, suffix), ', '.join(str(root_name2date[_]) for _ in root_names))

    edge_size = len(roots)
    if edge_size > 1:
        n.add_feature('node_meta{}'.format(suffix), 1)
    n.add_feature('{}{}'.format(EDGE_NAME, suffix), str(edge_size) if edge_size != 1 else '')
    e_size = e_size_scaling(transform_e_size(edge_size))
    n.add_feature('{}{}'.format(EDGE_SIZE, suffix), e_size)


def set_cyto_features_tree(n, state):
    n.add_feature(NODE_NAME, state)
    n.add_feature(EDGE_NAME, n.dist)


def _forest2json_compressed(forest, compressed_forest, columns, name_feature, get_date, milestones=None,
                            dates_are_dates=True, is_mixed=False):
    e_size_scaling, font_scaling, size_scaling, transform_e_size, transform_size = \
        get_size_transformations(compressed_forest)

    sort_key = lambda n: (getattr(n, UNRESOLVED, 0),
                          get_column_value_str(n, name_feature, format_list=True) if name_feature else '',
                          *(get_column_value_str(n, column, format_list=True) for column in columns),
                          -getattr(n, NUM_TIPS_INSIDE),
                          -len(getattr(n, ROOTS)),
                          n.name)
    i = 0
    node2id = {}
    todo = Queue()
    for compressed_tree in compressed_forest:
        todo.put_nowait(compressed_tree)
        node2id[compressed_tree] = i
        i += 1
        while not todo.empty():
            n = todo.get_nowait()
            for c in sorted(n.children, key=sort_key):
                node2id[c] = i
                i += 1
                todo.put_nowait(c)

    n2state = {}

    # Set the cytoscape features
    for compressed_tree in compressed_forest:
        for n in compressed_tree.traverse():
            state = get_column_value_str(n, name_feature, format_list=False, list_value='') if name_feature else ''
            n2state[n] = state
            root_names = [_.name for _ in getattr(n, ROOTS)]
            root_dates = [get_formatted_date(_, dates_are_dates) for _ in getattr(n, ROOTS)]
            set_cyto_features_compressed(n, size_scaling, e_size_scaling, font_scaling,
                                         transform_size, transform_e_size, state, root_names, root_dates,
                                         is_mixed=is_mixed)

    # Calculate node coordinates
    min_size = 2 * min(min(getattr(_, NODE_SIZE) for _ in compressed_tree.traverse())
                       for compressed_tree in compressed_forest)
    n2width = {}
    for compressed_tree in compressed_forest:
        for n in compressed_tree.traverse('postorder'):
            n2width[n] = max(getattr(n, NODE_SIZE),
                             sum(n2width[c] for c in n.children) + min_size * (len(n.children) - 1))

    n2x, n2y = {}, {compressed_tree: 0 for compressed_tree in compressed_forest}
    n2offset = {}
    tree_offset = 0
    for compressed_tree in compressed_forest:
        n2offset[compressed_tree] = tree_offset
        tree_offset += n2width[compressed_tree] + 2 * min_size
        for n in compressed_tree.traverse('preorder'):
            n2x[n] = n2offset[n] + n2width[n] / 2
            offset = n2offset[n]
            if not n.is_leaf():
                for c in sorted(n.children, key=lambda c: node2id[c]):
                    n2offset[c] = offset
                    offset += n2width[c] + min_size
                    n2y[c] = n2y[n] + getattr(n, NODE_SIZE) / 2 + getattr(c, NODE_SIZE) / 2 + min_size
        for n in compressed_tree.traverse('postorder'):
            if not n.is_leaf():
                n2x[n] = np.mean([n2x[c] for c in n.children])

    def filter_by_date(items, date):
        return [_ for _ in items if get_date(_) <= date]

    # Set the cytoscape feature for different timeline points
    for tree, compressed_tree in zip(forest, compressed_forest):
        if len(milestones) > 1:
            nodes = list(compressed_tree.traverse())
            for i in range(len(milestones) - 1, -1, -1):
                milestone = milestones[i]
                nodes_i = []

                # remove too recent nodes from the original tree
                for n in tree.traverse('postorder'):
                    if n.is_root():
                        continue
                    if get_date(n) > milestone:
                        n.up.remove_child(n)

                suffix = '_{}'.format(i)
                for n in nodes:
                    state = n2state[n]
                    tips_inside, tips_below, internal_nodes_inside, roots = getattr(n, TIPS_INSIDE, []), \
                                                                            getattr(n, TIPS_BELOW, []), \
                                                                            getattr(n, INTERNAL_NODES_INSIDE, []), \
                                                                            getattr(n, ROOTS, [])
                    tips_inside_i, tips_below_i, internal_nodes_inside_i, roots_i = [], [], [], []
                    for ti, tb, ini, root in zip(tips_inside, tips_below, internal_nodes_inside, roots):
                        if get_date(root) <= milestone:
                            roots_i.append(root)

                            ti = filter_by_date(ti, milestone)
                            tb = [_ for _ in tb if getattr(_, DATE) <= milestone]
                            ini = filter_by_date(ini, milestone)

                            tips_inside_i.append(ti + [_ for _ in ini if _.is_leaf()])
                            tips_below_i.append(tb)
                            internal_nodes_inside_i.append([_ for _ in ini if not _.is_leaf()])
                    n.add_features(**{TIPS_INSIDE: tips_inside_i, TIPS_BELOW: tips_below_i,
                                      INTERNAL_NODES_INSIDE: internal_nodes_inside_i, ROOTS: roots_i})
                    if roots_i:
                        n.add_feature(MILESTONE, i)
                        root_names = [getattr(_, BRANCH_NAME) if getattr(_, DATE) > milestone else _.name for _ in
                                      roots_i]
                        root_dates = [getattr(_, DATE) for _ in roots_i]
                        if dates_are_dates:
                            try:
                                root_dates = [numeric2datetime(_).strftime("%d %b %Y") for _ in root_dates]
                            except:
                                pass
                        set_cyto_features_compressed(n, size_scaling, e_size_scaling, font_scaling, transform_size,
                                                     transform_e_size, state, root_names=root_names, suffix=suffix,
                                                     root_dates=root_dates, is_mixed=is_mixed)
                        nodes_i.append(n)
                nodes = nodes_i

    # Save the structure
    clazzes = set()
    nodes, edges = [], []

    one_column = columns[0] if len(columns) == 1 else None

    for n, n_id in node2id.items():
        if one_column:
            values = getattr(n, one_column, set())
            clazz = tuple(sorted(values))
        else:
            clazz = tuple('{}_{}'.format(column, get_column_value_str(n, column, format_list=False, list_value=''))
                          for column in columns)
        if clazz:
            clazzes.add(clazz)
        nodes.append(get_node(n, n_id, tooltip=get_tooltip(n, columns),
                              clazz=clazz, x=n2x[n], y=n2y[n]))

        for child in sorted(n.children, key=lambda _: node2id[_]):
            edge_attributes = {feature: getattr(child, feature) for feature in child.features
                               if feature.startswith('edge_') or feature == MILESTONE or feature == IS_POLYTOMY}
            source_name = n_id
            edges.append(get_edge(source_name, node2id[child], **edge_attributes))

    json_dict = {NODES: nodes, EDGES: edges}
    return json_dict, sorted(clazzes)


def binary_search(start, end, value, array):
    if start >= end - 1:
        return start
    i = int((start + end) / 2)
    if array[i] == value or array[i] > value and (i == start or value > array[i - 1]):
        return i
    if array[i] > value:
        return binary_search(start, i, value, array)
    return binary_search(i + 1, end, value, array)


def _forest2json(forest, columns, name_feature, get_date, milestones=None, timeline_type=TIMELINE_SAMPLED,
                 dates_are_dates=True):
    min_root_date = min(getattr(tree, DATE) for tree in forest)
    width = sum(len(tree) for tree in forest)
    height_factor = 300 * width / max(
        (max(getattr(_, DATE) for _ in tree) - min_root_date + tree.dist) for tree in forest)
    zero_dist = min(min(min(_.dist for _ in tree.traverse() if _.dist > 0) for tree in forest), 300) * height_factor / 2

    # Calculate node coordinates
    n2x, n2y = {}, {}
    x = -600
    for tree in forest:
        x += 600
        for t in tree:
            n2x[t] = x
            x += 600

    for tree in forest:
        for n in tree.traverse('postorder'):
            state = get_column_value_str(n, name_feature, format_list=False, list_value='') if name_feature else ''
            n.add_feature('node_root_id', n.name)
            n.add_feature('node_root_date', get_formatted_date(n, dates_are_dates))
            if not n.is_leaf():
                n2x[n] = np.mean([n2x[_] for _ in n.children])
            n2y[n] = (getattr(n, DATE) - min_root_date) * height_factor
            for c in n.children:
                if n2y[c] == n2y[n]:
                    n2y[c] += zero_dist
            set_cyto_features_tree(n, state)

    # Save cytoscape features at different timeline points
    if len(milestones) > 1:
        for tree in forest:
            for n in tree.traverse('preorder'):
                ms_i = binary_search(0 if n.is_root() else getattr(n.up, MILESTONE),
                                     len(milestones), get_date(n), milestones)
                n.add_feature(MILESTONE, ms_i)
                for i in range(len(milestones) - 1, ms_i - 1, -1):
                    milestone = milestones[i]
                    suffix = '_{}'.format(i)
                    if TIMELINE_LTT == timeline_type:
                        # if it is LTT also cut the branches if needed
                        if getattr(n, DATE) > milestone:
                            n.add_feature('{}{}'.format(EDGE_NAME, suffix),
                                          np.round(milestone - getattr(n.up, DATE), 3))
                        else:
                            n.add_feature('{}{}'.format(EDGE_NAME, suffix), np.round(n.dist, 3))

    # Save the structure
    clazzes = set()
    nodes, edges = [], []

    one_column = columns[0] if len(columns) == 1 else None

    i = 0
    node2id = {}
    for tree in forest:
        for n in tree.traverse():
            node2id[n] = i
            i += 1

    for n, n_id in node2id.items():
        if n.is_root():
            fake_id = 'fake_node_{}'.format(n_id)
            nodes.append(get_fake_node(fake_id, n2x[n], n2y[n] - n.dist * height_factor))
            edges.append(get_edge(fake_id, n_id, **{feature: getattr(n, feature) for feature in n.features
                                                    if feature.startswith('edge_') or feature == MILESTONE}))
        if one_column:
            values = getattr(n, one_column, set())
            clazz = tuple(sorted(values))
        else:
            clazz = tuple('{}_{}'.format(column, get_column_value_str(n, column, format_list=False, list_value=''))
                          for column in columns)
        if clazz:
            clazzes.add(clazz)
        nodes.append(get_node(n, n_id, tooltip=get_tooltip(n, columns), clazz=clazz, x=n2x[n], y=n2y[n]))

        for child in n.children:
            edge_attributes = {feature: getattr(child, feature) for feature in child.features
                               if feature.startswith('edge_') or feature == MILESTONE or feature == IS_POLYTOMY}
            source_name = n_id
            target_name = 'fake_node_{}'.format(node2id[child])
            nodes.append(get_fake_node(target_name, x=n2x[child], y=n2y[n]))
            edges.append(get_edge(source_name, target_name, fake=1,
                                  **{k: v for (k, v) in edge_attributes.items()
                                     if EDGE_NAME not in k and IS_POLYTOMY not in k}))
            source_name = target_name
            edges.append(get_edge(source_name, node2id[child], **edge_attributes))

    json_dict = {NODES: nodes, EDGES: edges}
    return json_dict, sorted(clazzes)


def _forest2json_transitions(states, counts, transitions, state2colour, threshold=0):
    nodes, edges = [], []
    n = len(states)

    n_scaler = get_scaling_function(y_m=200, y_M=800, x_m=min(counts), x_M=max(counts))
    font_scaler = get_scaling_function(y_m=MIN_FONT_SIZE, y_M=MIN_FONT_SIZE * 3, x_m=min(counts), x_M=max(counts))
    positive_transitions = transitions[transitions > 0]
    e_scaler = get_scaling_function(y_m=30, y_M=200, x_m=min(positive_transitions), x_M=max(positive_transitions))

    transtions_to_from = np.transpose(transitions.copy())
    np.fill_diagonal(transtions_to_from, 0)
    nums = np.triu(transitions + transtions_to_from).flatten()
    positive_nums = nums[nums > 0]

    max_transition_num = max(positive_nums)
    if max_transition_num <= 2:
        miles = sorted((set({0, threshold, 1 if max_transition_num > 1 else 0}
                                     | set(np.round(positive_nums, 3))) - {np.round(max_transition_num, 3)}))
    else:
        miles = sorted(set({0, threshold, 1} | set(np.trunc(positive_nums))))
    miles = np.array([_ for _ in miles if threshold <= _ < max_transition_num])
    if not len(miles):
        miles = [0]

    # we hide connections when they are < mile
    def get_mile(start, end, value):
        if start == end:
            return None
        i = int((start + end) / 2)
        if miles[i] <= value:
            if i == end - 1 or value < miles[i + 1]:
                return i
            return get_mile(i + 1, end, value)
        return get_mile(start, i, value)

    i2mile = defaultdict(lambda: -1)
    for i in range(n):
        from_state = states[i]
        n_tips = counts[i]
        i_node_size = n_scaler(n_tips)
        if n_tips > 0:
            for j in range(i, n):
                if counts[j] > 0:
                    to_state = states[j]
                    n_ij = transitions[i, j]
                    n_ji = transitions[j, i]
                    node_size = (i_node_size + n_scaler(counts[j])) / 2
                    if n_ij > 0:
                        mile = get_mile(0, len(miles), n_ij)
                        if mile is not None:
                            i2mile[i] = max(mile, i2mile[i])
                            i2mile[j] = max(mile, i2mile[j])
                            edges.append(get_edge(i, j,
                                                  **{ID: '{}_{}'.format(i, j),
                                                     EDGE_SIZE: e_scaler(n_ij),
                                                     NODE_SIZE: node_size / (2 if i != j else 1),
                                                     EDGE_NAME: n_ij,
                                                     TOOLTIP: '{} transitions from {} to {}'
                                                  .format(n_ij, from_state, to_state),
                                                     MILESTONE: mile}))
                    if n_ji > 0 and i != j:
                        mile = get_mile(0, len(miles), n_ji)
                        if mile is not None:
                            i2mile[i] = max(mile, i2mile[i])
                            i2mile[j] = max(mile, i2mile[j])
                            edges.append(get_edge(j, i,
                                                  **{ID: '{}_{}'.format(j, i),
                                                     EDGE_SIZE: e_scaler(n_ji),
                                                     NODE_SIZE: node_size / 2,
                                                     EDGE_NAME: n_ji,
                                                     TOOLTIP: '{} transitions from {} to {}'
                                                  .format(n_ji, to_state, from_state),
                                                     MILESTONE: mile}))
            if i2mile[i] >= 0:
                nodes.append(_get_node(data={ID: i, NODE_NAME: '{} ({:.0f})'.format(from_state, n_tips),
                                             NODE_SIZE: i_node_size,
                                             FONT_SIZE: font_scaler(n_tips),
                                             TOOLTIP: '{} is represented by {:.0f} samples.'.format(from_state, n_tips),
                                             COLOUR: state2colour[from_state],
                                             MILESTONE: i2mile[i]}))
    json_dict = {NODES: nodes, EDGES: edges}
    return json_dict, ['{:g}'.format(_) for _ in miles]


def get_size_transformations(forest):
    max_size, min_size, max_e_size, min_e_size = 1, np.inf, 1, np.inf
    for tree in forest:
        for n in tree.traverse():
            sz = max(getattr(n, NUM_TIPS_INSIDE), 1)
            max_size = max(max_size, sz)
            min_size = min(min_size, sz)
            e_sz = len(getattr(n, ROOTS))
            max_e_size = max(max_e_size, e_sz)
            min_e_size = min(min_e_size, e_sz)

    need_log = max_size / min_size > 100
    transform_size = lambda _: np.power(np.log10(_ + 9) if need_log else _, 1 / 2)

    need_e_log = max_e_size / min_e_size > 100
    transform_e_size = lambda _: np.log10(_) if need_e_log else _

    size_scaling = get_scaling_function(y_m=MIN_NODE_SIZE, y_M=MIN_NODE_SIZE * min(8, int(max_size / min_size)),
                                        x_m=transform_size(min_size), x_M=transform_size(max_size))
    font_scaling = get_scaling_function(y_m=MIN_FONT_SIZE, y_M=MIN_FONT_SIZE * min(3, int(max_size / min_size)),
                                        x_m=transform_size(min_size), x_M=transform_size(max_size))
    e_size_scaling = get_scaling_function(y_m=MIN_EDGE_SIZE, y_M=MIN_EDGE_SIZE * min(3, int(max_e_size / min_e_size)),
                                          x_m=transform_e_size(min_e_size), x_M=transform_e_size(max_e_size))

    return e_size_scaling, font_scaling, size_scaling, transform_e_size, transform_size


def get_tooltip(n, columns):
    return '<br>'.join('{}: {}'.format(column, get_column_value_str(n, column, format_list=True))
                       for column in columns)


def save_as_cytoscape_html(forest, out_html, column2states, name_feature, name2colour, compressed_forest,
                           milestone_label, timeline_type, milestones, get_date, work_dir, local_css_js=False,
                           milestone_labels=None, is_mixed=False):
    """
    Converts a forest to an html representation using Cytoscape.js.

    If categories are specified they are visualised as pie-charts inside the nodes,
    given that each node contains features corresponding to these categories with values being the percentage.
    For instance, given categories ['A', 'B', 'C'], a node with features {'A': 50, 'B': 50}
    will have a half-half pie-chart (half-colored in a colour of A, and half B).

    :param name_feature: str, a node feature whose value will be used as a label
    :param name2colour: dict, str to str, category name to HEX colour mapping 
    :param forest: list(ete3.Tree)
    :param out_html: path where to save the resulting html file.
    """
    graph_name = os.path.splitext(os.path.basename(out_html))[0]
    columns = sorted(column2states.keys())
    if milestone_labels is None:
        milestone_labels = ['{:g}'.format(_) for _ in milestones]

    if compressed_forest is not None:
        json_dict, clazzes \
            = _forest2json_compressed(forest, compressed_forest, columns, name_feature=name_feature, get_date=get_date,
                                      milestones=milestones, dates_are_dates=milestone_label == DATE_LABEL,
                                      is_mixed=is_mixed)
    else:
        json_dict, clazzes \
            = _forest2json(forest, columns, name_feature=name_feature, get_date=get_date, milestones=milestones,
                           timeline_type=timeline_type, dates_are_dates=milestone_label == DATE_LABEL)
    loader = PackageLoader('pastml')
    env = Environment(loader=loader)
    template = env.get_template('pie_tree.js') if compressed_forest is not None \
        else env.get_template('pie_tree_simple.js')

    clazz2css = {}
    for clazz_list in clazzes:
        n = len(clazz_list)
        css = ''
        for i, cat in enumerate(clazz_list, start=1):
            css += """
                'pie-{i}-background-color': "{colour}",
                'pie-{i}-background-size': '{percent}\%',
            """.format(i=i, percent=round(100 / n, 2), colour=name2colour[cat])
        clazz2css[_clazz_list2css_class(clazz_list)] = css
    graph = template.render(clazz2css=clazz2css.items(), elements=json_dict, title=graph_name,
                            years=milestone_labels,
                            tips='samples' if TIMELINE_SAMPLED == timeline_type
                            else ('lineages ending' if TIMELINE_LTT == timeline_type else 'external nodes'),
                            internal_nodes='internal nodes' if TIMELINE_NODES == timeline_type
                            else 'diversification events',
                            age_label=milestone_label)
    slider = env.get_template('time_slider.html').render(min_date=0, max_date=len(milestones) - 1,
                                                         cur_date=len(milestones) - 1,
                                                         name=milestone_label) if len(milestones) > 1 else ''

    template = env.get_template('index.html')
    os.makedirs(os.path.abspath(os.path.dirname(out_html)), exist_ok=True)

    if local_css_js:
        js_list = []
        os.makedirs(os.path.join(work_dir, 'js'), exist_ok=True)
        os.makedirs(os.path.join(work_dir, 'css'), exist_ok=True)
        os.makedirs(os.path.join(work_dir, 'fonts'), exist_ok=True)

        template_dir = os.path.join(os.path.abspath(os.path.split(__file__)[0]), '..', 'templates')
        for _ in sorted(glob(os.path.join(template_dir, 'js', '*.js*'))):
            cp = os.path.join(work_dir, 'js', os.path.split(_)[1])
            copyfile(_, cp)
            if cp.endswith('.js'):
                js_list.append(cp)
        css_list = []
        for _ in glob(os.path.join(template_dir, 'css', '*.css*')):
            cp = os.path.join(work_dir, 'css', os.path.split(_)[1])
            copyfile(_, cp)
            if cp.endswith('.css'):
                css_list.append(cp)
        for _ in glob(os.path.join(template_dir, 'fonts', '*.*')):
            cp = os.path.join(work_dir, 'fonts', os.path.split(_)[1])
            copyfile(_, cp)
    else:
        js_list = JS_LIST
        css_list = CSS_LIST
    page = template.render(graph=graph, title=graph_name, slider=slider, js_list=js_list, css_list=css_list)

    with open(out_html, 'w+') as fp:
        fp.write(page)


def save_as_transition_html(character, states, counts, transitions, out_html, state2colour, work_dir,
                            local_css_js=False, threshold=0):
    """
    Converts transition count data to an html representation using Cytoscape.js.

    :param out_html: path where to save the resulting html file.
    """
    graph_name = os.path.splitext(os.path.basename(out_html))[0]
    transitions[transitions < threshold] = 0
    json_dict, thresholds = _forest2json_transitions(states, counts, transitions, state2colour, threshold=threshold)

    loader = PackageLoader('pastml')
    env = Environment(loader=loader)
    template = env.get_template('transitions.js')

    graph = template.render(elements=json_dict, character=character,
                            years=thresholds)
    slider = env.get_template('time_slider.html').render(min_date=0, max_date=len(thresholds) - 1,
                                                         name='transition number threshold', cur_date=0) if len(
        thresholds) > 1 else ''

    template = env.get_template('index.html')
    os.makedirs(os.path.abspath(os.path.dirname(out_html)), exist_ok=True)

    if local_css_js:
        js_list = []
        os.makedirs(os.path.join(work_dir, 'js'), exist_ok=True)
        os.makedirs(os.path.join(work_dir, 'css'), exist_ok=True)
        os.makedirs(os.path.join(work_dir, 'fonts'), exist_ok=True)

        template_dir = os.path.join(os.path.abspath(os.path.split(__file__)[0]), '..', 'templates')
        for _ in sorted(glob(os.path.join(template_dir, 'js', '*.js*'))):
            cp = os.path.join(work_dir, 'js', os.path.split(_)[1])
            copyfile(_, cp)
            if cp.endswith('.js'):
                js_list.append(cp)
        css_list = []
        for _ in glob(os.path.join(template_dir, 'css', '*.css*')):
            cp = os.path.join(work_dir, 'css', os.path.split(_)[1])
            copyfile(_, cp)
            if cp.endswith('.css'):
                css_list.append(cp)
        for _ in glob(os.path.join(template_dir, 'fonts', '*.*')):
            cp = os.path.join(work_dir, 'fonts', os.path.split(_)[1])
            copyfile(_, cp)
    else:
        js_list = JS_LIST
        css_list = CSS_LIST
    page = template.render(graph=graph, title=graph_name, slider=slider, js_list=js_list, css_list=css_list)

    with open(out_html, 'w+') as fp:
        fp.write(page)


def _clazz_list2css_class(clazz_list):
    if not clazz_list:
        return None
    return ''.join(c for c in '-'.join(clazz_list) if c.isalnum() or '-' == c)


def _get_node(data, clazz=None, position=None):
    if position:
        data['node_x'] = float(position[0])
        data['node_y'] = float(position[1])
    res = {DATA: data}
    if clazz:
        res['classes'] = clazz
    return res


def _get_edge(**data):
    return {DATA: data}


def get_column_value_str(n, column, format_list=True, list_value=''):
    values = getattr(n, column, set())
    if isinstance(values, str):
        return values
    return ' or '.join(sorted(values)) if format_list or len(values) == 1 else list_value


def visualize(forest, column2states, work_dir, name_column=None, html=None, html_compressed=None, html_mixed=None,
              tip_size_threshold=REASONABLE_NUMBER_OF_TIPS, date_label='Dist. to root', timeline_type=TIMELINE_SAMPLED,
              local_css_js=False, column2colours=None, focus=None, pajek=None, pajek_timing=VERTICAL):
    for tree in forest:
        nodes_in_focus = set()
        for node in tree.traverse():
            for column in column2states.keys():
                col_state = getattr(node, column, set())
                if focus and col_state & focus[column]:
                    nodes_in_focus.add(node)
        for node in nodes_in_focus:
            node.add_feature(IN_FOCUS, True)
            if not node.is_root() and node.up not in nodes_in_focus:
                node.up.add_feature(UP_FOCUS, True)
            for c in node.children:
                if c not in nodes_in_focus:
                    c.add_feature(AROUND_FOCUS, True)

    one_column = next(iter(column2states.keys())) if len(column2states) == 1 else None

    name2colour = {}
    for column, states in column2states.items():
        num_unique_values = len(states)
        colours = None
        if column2colours and column in column2colours:
            try:
                colours = parse_colours(column2colours[column], states)
            except ValueError as e:
                logging.getLogger('pastml').error('Failed to parse the input colours: {}'.format(e))
        if colours is None:
            colours = get_enough_colours(num_unique_values)
        for value, col in zip(states, colours):
            name2colour[value if one_column else '{}_{}'.format(column, value)] = col
        state2color = dict(zip(states, colours))
        # let ambiguous values be white
        if one_column is None:
            name2colour['{}_'.format(column)] = WHITE
        if column == name_column:
            for tree in forest:
                for n in tree.traverse():
                    sts = getattr(n, column, set())
                    if len(sts) == 1 and not n.is_root() and getattr(n.up, column, set()) == sts:
                        n.add_feature('edge_color', state2color[next(iter(sts))])
        out_colour_file = os.path.join(work_dir, get_pastml_colour_file(column))
        # Not using DataFrames to speed up document writing
        with open(out_colour_file, 'w+') as f:
            f.write('state\tcolour\n')
            for s in sorted(states):
                f.write('{}\t{}\n'.format(s, state2color[s]))
        logging.getLogger('pastml').debug('Mapped states to colours for {} as following: {} -> {}, '
                                          'and serialized this mapping to {}.'
                                          .format(column, states, colours, out_colour_file))
    for tree in forest:
        for node in tree.traverse():
            if node.is_leaf():
                node.add_feature(IS_TIP, True)
            node.add_feature(BRANCH_NAME, '{}-{}'.format(node.up.name if not node.is_root() else '', node.name))
            for column in column2states.keys():
                col_state = getattr(node, column, set())
                if len(col_state) != 1:
                    node.add_feature(UNRESOLVED, 1)
                    break

    if TIMELINE_NODES == timeline_type:
        def get_date(node):
            return getattr(node, DATE)
    elif TIMELINE_SAMPLED == timeline_type:
        max_date = max(max(getattr(_, DATE) for _ in tree) for tree in forest)

        def get_date(node):
            tips = [_ for _ in node if getattr(_, IS_TIP, False)]
            return min(getattr(_, DATE) for _ in tips) if tips else max_date
    elif TIMELINE_LTT == timeline_type:
        def get_date(node):
            return getattr(node, DATE) if node.is_root() else (getattr(node.up, DATE) + 1e-6)
    else:
        raise ValueError('Unknown timeline type: {}. Allowed ones are {}, {} and {}.'
                         .format(timeline_type, TIMELINE_NODES, TIMELINE_SAMPLED, TIMELINE_LTT))
    dates = []
    for tree in forest:
        dates.extend([getattr(_, DATE) for _ in (tree.traverse()
                                                 if timeline_type in [TIMELINE_LTT, TIMELINE_NODES] else tree)])
    dates = sorted(dates)
    milestones = sorted({dates[0], dates[len(dates) // 8], dates[len(dates) // 4], dates[3 * len(dates) // 8],
                         dates[len(dates) // 2], dates[5 * len(dates) // 8], dates[3 * len(dates) // 4],
                         dates[7 * len(dates) // 8], dates[-1]})
    milestone_labels = None
    if DATE_LABEL == date_label:
        try:
            milestone_labels = [numeric2datetime(_).strftime("%d %b %Y") for _ in milestones]
        except:
            pass
    if milestone_labels is None:
        milestone_labels = ['{:g}'.format(_) for _ in milestones]

    if html:
        total_num_tips = sum(len(tree) for tree in forest)
        if total_num_tips > MAX_TIPS_FOR_FULL_TREE_VISUALISATION:
            logging.getLogger('pastml').error('The full tree{} will not be visualised as {} too large ({} tips): '
                                              'the limit is {} tips. Check out upload to iTOL option instead.'
                                              .format('s' if len(forest) > 1 else '',
                                                      'they are' if len(forest) > 1 else 'it is',
                                                      total_num_tips, MAX_TIPS_FOR_FULL_TREE_VISUALISATION))
        else:
            save_as_cytoscape_html(forest, html, column2states=column2states, name2colour=name2colour,
                                   name_feature='name', compressed_forest=None, milestone_label=date_label,
                                   timeline_type=timeline_type, milestones=milestones, get_date=get_date,
                                   work_dir=work_dir, local_css_js=local_css_js, milestone_labels=milestone_labels)
    if html_compressed and html_mixed:
        forest_mixed = copy_forest(forest)
    else:
        forest_mixed = forest

    if html_compressed:
        pajek_vert_arcs = [[], []] if pajek else None
        compressed_forest = [compress_tree(tree, columns=column2states.keys(), tip_size_threshold=tip_size_threshold,
                                           mixed=False, pajek=pajek_vert_arcs, pajek_timing=pajek_timing)
                             for tree in forest]
        if pajek:
            save_to_pajek(*pajek_vert_arcs, pajek)

        milestone_labels, milestones = update_milestones(forest, date_label, milestone_labels, milestones,
                                                         timeline_type)

        save_as_cytoscape_html(forest, html_compressed,
                               column2states=column2states, name2colour=name2colour,
                               name_feature=name_column, compressed_forest=compressed_forest,
                               milestone_label=date_label, timeline_type=timeline_type,
                               milestones=milestones, get_date=get_date, work_dir=work_dir, local_css_js=local_css_js,
                               milestone_labels=milestone_labels, is_mixed=False)

    if html_mixed:
        mixed_forest = [compress_tree(tree, columns=column2states.keys(), tip_size_threshold=tip_size_threshold,
                                      mixed=True) for tree in forest_mixed]
        milestone_labels, milestones = update_milestones(forest_mixed, date_label, milestone_labels, milestones,
                                                         timeline_type)
        save_as_cytoscape_html(forest_mixed, html_mixed,
                               column2states=column2states, name2colour=name2colour,
                               name_feature=name_column, compressed_forest=mixed_forest,
                               milestone_label=date_label, timeline_type=timeline_type,
                               milestones=milestones, get_date=get_date, work_dir=work_dir, local_css_js=local_css_js,
                               milestone_labels=milestone_labels, is_mixed=True)


def update_milestones(forest, date_label, milestone_labels, milestones, timeline_type):
    # If we trimmed a few tips while compressing and they happened to be the oldest/newest ones,
    # we should update the milestones accordingly.
    first_date, last_date = np.inf, -np.inf
    for tree in forest:
        for _ in (tree.traverse() if timeline_type in [TIMELINE_LTT, TIMELINE_NODES] else tree):
            first_date = min(first_date, getattr(_, DATE))
            last_date = max(last_date, getattr(_, DATE))
    milestones = [ms for ms in milestones if first_date <= ms <= last_date]
    if milestones[0] > first_date:
        milestones.insert(0, first_date)
    if milestones[-1] < last_date:
        milestones.append(last_date)
    if DATE_LABEL == date_label:
        try:
            milestone_labels = [numeric2datetime(_).strftime("%d %b %Y") for _ in milestones]
        except:
            pass
    if milestone_labels is None:
        milestone_labels = ['{:g}'.format(_) for _ in milestones]
    return milestone_labels, milestones



import logging
from datetime import datetime

PASTML_VERSION = '1.9.50'

METHOD = 'method'
STATES = 'states'
CHARACTER = 'character'

NUM_SCENARIOS = 'num_scenarios'
NUM_UNRESOLVED_NODES = 'num_unresolved_nodes'
NUM_STATES_PER_NODE = 'num_states_per_node_avg'
PERC_UNRESOLVED = 'percentage_of_unresolved_nodes'
NUM_NODES = 'num_nodes'
NUM_TIPS = 'num_tips'


def datetime2numeric(d):
    """
    Converts a datetime date to numeric format.
    For example: 2016-12-31 -> 2016.9972677595629; 2016-1-1 -> 2016.0
    :param d: a date to be converted
    :type d: np.datetime
    :return: numeric representation of the date
    :rtype: float
    """
    first_jan_this_year = datetime(year=d.year, month=1, day=1)
    day_of_this_year = d - first_jan_this_year
    first_jan_next_year = datetime(year=d.year + 1, month=1, day=1)
    days_in_this_year = first_jan_next_year - first_jan_this_year
    return d.year + day_of_this_year / days_in_this_year


def numeric2datetime(d):
    """
    Converts a numeric date to  datetime format.
    For example: 2016.9972677595629 -> 2016-12-31; 2016.0 ->  2016-1-1
    :param d: numeric representation of a date to be converted
    :type d: float
    :return: the converted date
    :rtype: np.datetime
    """
    year = int(d)
    first_jan_this_year = datetime(year=year, month=1, day=1)
    first_jan_next_year = datetime(year=year + 1, month=1, day=1)
    days_in_this_year = first_jan_next_year - first_jan_this_year
    day_of_this_year = int(round(days_in_this_year.days * (d % 1), 6)) + 1
    for m in range(1, 13):
        days_in_m = (datetime(year=year if m < 12 else (year + 1), month=m % 12 + 1, day=1)
                     - datetime(year=year, month=m, day=1)).days
        if days_in_m >= day_of_this_year:
            return datetime(year=year, month=m, day=day_of_this_year)
        day_of_this_year -= days_in_m


def col_name2cat(column):
    """
    Reformats the column string to make sure it contains only numerical, letter characters or underscore.

    :param column: column name to be reformatted
    :type column: str
    :return: column name with illegal characters removed
    :rtype: str
    """
    column_string = ''.join(s for s in column.replace(' ', '_') if s.isalnum() or '_' == s)
    return column_string


def get_personalized_feature_name(character, feature):
    """
    Precedes the feature name by the character name
    (useful when likelihoods for different characters are calculated in parallel).

    :param character: str, character name
    :param feature: str, feature to be personalized
    :return: str, the personalized feature
    """
    return '{}_{}'.format(character, feature)


def value2list(n, value, default_value):
    # create a variable for n columns
    # specifying the default value if nothing was chosen
    if value is None:
        value = default_value
    # and propagating the chosen value to all columns
    if not isinstance(value, list):
        value = [value] * n
    elif len(value) == 1:
        value = value * n
    # or making sure that the default value is chosen for the columns for which the value was not specified
    else:
        value += [default_value] * (n - len(value))
    return value


def _set_up_pastml_logger(verbose):
    logger = logging.getLogger('pastml')
    logger.setLevel(level=logging.DEBUG if verbose else logging.ERROR)
    logger.propagate = False
    if not logger.hasHandlers():
        ch = logging.StreamHandler()
        formatter = logging.Formatter('%(name)s:%(levelname)s:%(asctime)s %(message)s', datefmt="%H:%M:%S")
        ch.setFormatter(formatter)
        logger.addHandler(ch)
    return logger



import logging
import os
import re
from collections import Counter, defaultdict
from datetime import datetime

from Bio import Phylo
from ete3 import Tree, TreeNode
from ete3.parser.newick import NewickError
import pandas as pd
from pastml import datetime2numeric, PASTML_VERSION, _set_up_pastml_logger

POSTORDER = 'postorder'

INORDER = 'inorder'

PREORDER = 'preorder'

DATE = 'date'
DATE_CI = 'date_CI'

DATE_REGEX = r'[+-]*[\d]+[.\d]*(?:[e][+-][\d]+){0,1}'
DATE_COMMENT_REGEX = '[&,:]date[=]["]{{0,1}}({})["]{{0,1}}'.format(DATE_REGEX)
CI_DATE_REGEX_LSD = '[&,:]CI_date[=]["]{{0,1}}[{{]{{0,1}}({})\s*[,;]{{0,1}}\s*({})[}}]{{0,1}}["]{{0,1}}'.format(
    DATE_REGEX, DATE_REGEX)
CI_DATE_REGEX_PASTML = '[&,:]date_CI[=]["]{{0,1}}({})[|]({})["]{{0,1}}'.format(DATE_REGEX, DATE_REGEX)
COLUMN_REGEX_PASTML = '[&,]{column}[=]([^]^,]+)'

IS_POLYTOMY = 'polytomy'


def parse_date(d):
    try:
        return float(d)
    except ValueError:
        try:
            return datetime2numeric(pd.to_datetime(d, infer_datetime_format=True))
        except ValueError:
            raise ValueError('Could not infer the date format for root date "{}", please check it.'.format(d))


def get_dist_to_root(tip):
    dist_to_root = 0
    n = tip
    while not n.is_root():
        dist_to_root += n.dist
        n = n.up
    return dist_to_root


def annotate_dates(forest, root_dates=None, annotate_zeros=True):
    if root_dates is None:
        root_dates = [None] * len(forest)
    for tree, root_date in zip(forest, root_dates):
        for node in tree.traverse('preorder'):
            if getattr(node, DATE, None) is None:
                if node.is_root():
                    if root_date is not None or annotate_zeros:
                        node.add_feature(DATE, root_date if root_date else 0)
                else:
                    parent_date = getattr(node.up, DATE, None)
                    if parent_date is not None:
                        node.add_feature(DATE, parent_date + node.dist)
            else:
                node.add_feature(DATE, float(getattr(node, DATE)))
            ci = getattr(node, DATE_CI, None)
            if ci and not isinstance(ci, list) and not isinstance(ci, tuple):
                node.del_feature(DATE_CI)
                if isinstance(ci, str) and '|' in ci:
                    try:
                        node.add_feature(DATE_CI, [float(_) for _ in ci.split('|')])
                    except:
                        pass


def name_tree(tree, suffix=""):
    """
    Names all the tree nodes that are not named or have non-unique names, with unique names.

    :param tree: tree to be named
    :type tree: ete3.Tree

    :return: void, modifies the original tree
    """
    existing_names = Counter()
    n_nodes = 0
    for _ in tree.traverse():
        n_nodes += 1
        if _.name:
            existing_names[_.name] += 1
            if '.polytomy_' in _.name:
                _.add_feature(IS_POLYTOMY, 1)
    if n_nodes == len(existing_names):
        return
    i = 0
    new_existing_names = Counter()
    for node in tree.traverse('preorder'):
        name_prefix = node.name if node.name and existing_names[node.name] < 10 \
            else '{}{}{}'.format('t' if node.is_leaf() else 'n', i, suffix)
        name = 'root{}'.format(suffix) if node.is_root() else name_prefix
        while name is None or name in new_existing_names:
            name = '{}{}{}'.format(name_prefix, i, suffix)
            i += 1
        node.name = name
        new_existing_names[name] += 1


def collapse_zero_branches(forest, features_to_be_merged=None):
    """
    Collapses zero branches in tre tree/forest.

    :param forest: tree or list of trees
    :type forest: ete3.Tree or list(ete3.Tree)
    :param features_to_be_merged: list of features whose values are to be merged
        in case the nodes are merged during collapsing
    :type features_to_be_merged: list(str)
    :return: void
    """
    num_collapsed = 0

    if features_to_be_merged is None:
        features_to_be_merged = []

    for tree in forest:
        for n in list(tree.traverse('postorder')):
            zero_children = [child for child in n.children if not child.is_leaf() and child.dist <= 0]
            if not zero_children:
                continue
            for feature in features_to_be_merged:
                feature_intersection = set.intersection(*(getattr(child, feature, set()) for child in zero_children)) \
                                       & getattr(n, feature, set())
                if feature_intersection:
                    value = feature_intersection
                else:
                    value = set.union(*(getattr(child, feature, set()) for child in zero_children)) \
                            | getattr(n, feature, set())
                if value:
                    n.add_feature(feature, value)
            for child in zero_children:
                n.remove_child(child)
                for grandchild in child.children:
                    n.add_child(grandchild)
            num_collapsed += len(zero_children)
    if num_collapsed:
        logging.getLogger('pastml').debug('Collapsed {} internal zero branches.'.format(num_collapsed))


def remove_certain_leaves(tr, to_remove=lambda node: False):
    """
    Removes all the branches leading to leaves identified positively by to_remove function.
    :param tr: the tree of interest (ete3 Tree)
    :param to_remove: a method to check is a leaf should be removed.
    :return: void, modifies the initial tree.
    """

    tips = [tip for tip in tr if to_remove(tip)]
    for node in tips:
        if node.is_root():
            return None
        parent = node.up
        parent.remove_child(node)
        # If the parent node has only one child now, merge them.
        if len(parent.children) == 1:
            brother = parent.children[0]
            brother.dist += parent.dist
            if parent.is_root():
                brother.up = None
                tr = brother
            else:
                grandparent = parent.up
                grandparent.remove_child(parent)
                grandparent.add_child(brother)
    return tr


def read_forest(tree_path, columns=None):
    roots = None
    try:
        roots = parse_nexus(tree_path, columns=columns)
    except:
        pass
    if not roots:
        with open(tree_path, 'r') as f:
            nwks = f.read().replace('\n', '').split(';')
        if not nwks:
            raise ValueError('Could not find any trees (in newick or nexus format) in the file {}.'.format(tree_path))
        roots = [read_tree(nwk + ';', columns) for nwk in nwks[:-1]]

    num_neg = 0
    for root in roots:
        for _ in root.traverse():
            if _.dist < 0:
                num_neg += 1
                _.dist = 0
    if num_neg:
        logging.getLogger('pastml').warning('Input tree{} contained {} negative branches: we put them to zero.'
                                            .format('s' if len(roots) > 0 else '', num_neg))
    logging.getLogger('pastml').debug('Read the tree{} {}.'.format('s' if len(roots) > 0 else '', tree_path))
    return roots


def read_tree(tree_path, columns=None):
    tree = None
    for f in (3, 2, 5, 0, 1, 4, 6, 7, 8, 9):
        try:
            tree = Tree(tree_path, format=f)
            break
        except NewickError:
            try:
                tree = Tree(tree_path, format=f, quoted_node_names=True)
                break
            except NewickError:
                continue
    if not tree:
        raise ValueError('Could not read the tree {}. Is it a valid newick?'.format(tree_path))
    if columns:
        for n in tree.traverse():
            for c in columns:
                vs = set(getattr(n, c).split('|')) if hasattr(n, c) else set()
                if vs:
                    n.add_feature(c, vs)
    return tree


def parse_nexus(tree_path, columns=None):
    trees = []
    for nex_tree in read_nexus(tree_path):
        todo = [(nex_tree.root, None)]
        tree = None
        while todo:
            clade, parent = todo.pop()
            dist = 0
            try:
                dist = float(clade.branch_length)
            except:
                pass
            name = getattr(clade, 'name', None)
            if not name:
                name = getattr(clade, 'confidence', None)
                if not isinstance(name, str):
                    name = None
            node = TreeNode(dist=dist, name=name)
            if parent is None:
                tree = node
            else:
                parent.add_child(node)

            # Parse LSD2 dates and CIs, and PastML columns
            date, ci = None, None
            columns2values = defaultdict(set)
            comment = getattr(clade, 'comment', None)
            if isinstance(comment, str):
                date = next(iter(re.findall(DATE_COMMENT_REGEX, comment)), None)
                ci = next(iter(re.findall(CI_DATE_REGEX_LSD, comment)), None)
                if ci is None:
                    ci = next(iter(re.findall(CI_DATE_REGEX_PASTML, comment)), None)
                if columns:
                    for column in columns:
                        values = \
                            set.union(*(set(_.split('|')) for _ in re.findall(COLUMN_REGEX_PASTML.format(column=column),
                                                                              comment)), set())
                        if values:
                            columns2values[column] |= values
            comment = getattr(clade, 'branch_length', None)
            if not ci and not parent and isinstance(comment, str):
                ci = next(iter(re.findall(CI_DATE_REGEX_LSD, comment)), None)
                if ci is None:
                    ci = next(iter(re.findall(CI_DATE_REGEX_PASTML, comment)), None)
            comment = getattr(clade, 'confidence', None)
            if ci is None and comment is not None and isinstance(comment, str):
                ci = next(iter(re.findall(CI_DATE_REGEX_LSD, comment)), None)
                if ci is None:
                    ci = next(iter(re.findall(CI_DATE_REGEX_PASTML, comment)), None)
            if date is not None:
                try:
                    date = float(date)
                    node.add_feature(DATE, date)
                except:
                    pass
            if ci is not None:
                try:
                    ci = [float(_) for _ in ci]
                    node.add_feature(DATE_CI, ci)
                except:
                    pass
            if columns2values:
                for c, vs in columns2values.items():
                    node.add_feature(c, vs)
            todo.extend((c, node) for c in clade.clades)
        for n in tree.traverse('preorder'):
            date, ci = getattr(n, DATE, None), getattr(n, DATE_CI, None)
            if date is not None or ci is not None:
                for c in n.children:
                    if c.dist == 0:
                        if getattr(c, DATE, None) is None:
                            c.add_feature(DATE, date)
                        if getattr(c, DATE_CI, None) is None:
                            c.add_feature(DATE_CI, ci)
        for n in tree.traverse('postorder'):
            date, ci = getattr(n, DATE, None), getattr(n, DATE_CI, None)
            if not n.is_root() and n.dist == 0 and (date is not None or ci is not None):
                if getattr(n.up, DATE, None) is None:
                    n.up.add_feature(DATE, date)
                if getattr(n.up, DATE_CI, None) is None:
                    n.up.add_feature(DATE_CI, ci)

        # propagate dates up to the root if needed
        if getattr(tree, DATE, None) is None:
            dated_node = next((n for n in tree.traverse() if getattr(n, DATE, None) is not None), None)
            if dated_node:
                while dated_node != tree:
                    if getattr(dated_node.up, DATE, None) is None:
                        dated_node.up.add_feature(DATE, getattr(dated_node, DATE) - dated_node.dist)
                    dated_node = dated_node.up

        trees.append(tree)
    return trees


def read_nexus(tree_path):
    with open(tree_path, 'r') as f:
        nexus = f.read()
    # replace CI_date="2019(2018,2020)" with CI_date="2018 2020"
    nexus = re.sub(r'CI_date="({})\(({}),({})\)"'.format(DATE_REGEX, DATE_REGEX, DATE_REGEX), r'CI_date="\2 \3"',
                   nexus)
    temp = tree_path + '.{}.temp'.format(datetime.timestamp(datetime.now()))
    with open(temp, 'w') as f:
        f.write(nexus)
    trees = list(Phylo.parse(temp, 'nexus'))
    os.remove(temp)
    return trees


def depth_first_traversal(node):
    yield node, PREORDER
    for i, child in enumerate(node.children):
        if i != 0:
            yield node, INORDER
        for _ in depth_first_traversal(child):
            yield _
    yield node, POSTORDER


def resolve_trees(column2states, forest):
    """
    Resolved polytomies based on state predictions:
    if a parent P in a state A has n children 2 <= m < n of which are in state B,
    we add a parent to these children (who becomes a child of P) in a state B
    at the distance of the oldest of the children.

    :param column2states: character to possible state mapping
    :type column2states: dict
    :param forest: a forest of trees of interest
    :type list(ete.Tree)
    :return: number of newly created nodes.
    :rtype: int
    """
    columns = sorted(column2states.keys())

    col2state2i = {c: dict(zip(states, range(len(states)))) for (c, states) in column2states.items()}

    def get_prediction(n):
        return '.'.join('-'.join(str(i) for i in sorted([col2state2i[c][_] for _ in getattr(n, c, set())]))
                        for c in columns)

    num_new_nodes = 0

    for tree in forest:
        todo = [tree]
        while todo:
            n = todo.pop()
            todo.extend(n.children)
            if len(n.children) > 2:
                state2children = defaultdict(list)
                for c in n.children:
                    state2children[get_prediction(c)].append(c)
                if len(state2children) > 1:
                    for state, children in state2children.items():
                        if group_children_if_needed(n, children, columns, state):
                            num_new_nodes += 1
    if num_new_nodes:
        logging.getLogger('pastml').debug(
            'Created {} new internal nodes while resolving polytomies'.format(num_new_nodes))
    else:
        logging.getLogger('pastml').debug('Could not resolve any polytomy')
    return num_new_nodes


def states_are_different(n1, n2, columns):
    for c in columns:
        if not getattr(n1, c, set()) & getattr(n2, c, set()):
            return True
    return False


def group_children_if_needed(n, children, columns, state):
    if len(children) <= 1:
        return False
    child = min(children, key=lambda _: _.dist)
    if not states_are_different(n, child, columns):
        return False
    dist = child.dist
    pol = n.add_child(dist=dist, name='{}.polytomy_{}'.format(n.name, state))
    pol.add_feature(IS_POLYTOMY, 1)
    c_date = getattr(child, DATE)
    pol.add_feature(DATE, c_date)
    n_ci = getattr(n, DATE_CI, None)
    c_ci = getattr(child, DATE_CI, None)
    pol.add_feature(DATE_CI, (None if not n_ci or not isinstance(n_ci, list)
                              else [n_ci[0],
                                    (c_ci[1] if c_ci and isinstance(c_ci, list) and len(c_ci) > 1
                                     else c_date)]))
    for c in columns:
        pol.add_feature(c, getattr(child, c))
    for c in children:
        n.remove_child(c)
        pol.add_child(c, dist=c.dist - dist)
    return True


def unresolve_trees(column2states, forest):
    """
    Unresolves polytomies whose states do not correspond to child states after likelihood recalculation.

    :param column2states: character to possible state mapping
    :type column2states: dict
    :param forest: a forest of trees of interest
    :type list(ete.Tree)
    :return: number of newly deleted nodes.
    :rtype: int
    """
    columns = sorted(column2states.keys())

    col2state2i = {c: dict(zip(states, range(len(states)))) for (c, states) in column2states.items()}

    def get_prediction(n):
        return '.'.join('-'.join(str(i) for i in sorted([col2state2i[c][_] for _ in getattr(n, c, set())]))
                        for c in columns)

    num_removed_nodes = 0
    num_new_nodes = 0

    def remove_node(n):
        parent = n.up
        for c in n.children:
            parent.add_child(c, dist=c.dist + n.dist)
        parent.remove_child(n)

    num_polytomies = 0

    for tree in forest:
        for n in tree.traverse('postorder'):
            if getattr(n, IS_POLYTOMY, False):
                num_polytomies += 1

                state2children = defaultdict(list)
                n_children = list(n.children)
                for c in n_children:
                    state2children[get_prediction(c)].append(c)
                parent = n.up

                # if the state is the same as all the child states, it's still a good polytomy resolution
                if len(state2children) == 1 and not states_are_different(n, n_children[0], columns):
                    # Just need to check that it is not the same state as the parent (then we don't need this polytomy)
                    if not states_are_different(n, parent, columns):
                        num_removed_nodes += 1
                        remove_node(n)
                    continue

                num_removed_nodes += 1
                remove_node(n)

                # now let's try to create new polytomies above
                above_state2children = defaultdict(list)
                for c in parent.children:
                    state2children[get_prediction(c)].append(c)
                for state, children in above_state2children.items():
                    if len(children) <= 1:
                        continue
                    if set(children) != set(n_children):
                        if group_children_if_needed(parent, children, columns, state):
                            num_new_nodes += 1
    if num_removed_nodes:
        logging.getLogger('pastml').debug(
            'Removed {} polytomy resolution{} as inconsistent with model parameters.'
            .format(num_removed_nodes, 's' if num_removed_nodes > 1 else ''))
        if num_new_nodes:
            logging.getLogger('pastml').debug(
                'Created {} new polytomy resolution{}.'.format(num_new_nodes, 's' if num_new_nodes > 1 else ''))
    elif num_polytomies - num_removed_nodes + num_new_nodes:
        logging.getLogger('pastml').debug('All the polytomy resolutions are consistent with model parameters.')
    return num_removed_nodes


def clear_extra_features(forest, features):
    features = set(features) | {'name', 'dist', 'support'}
    for tree in forest:
        for n in tree.traverse():
            for f in set(n.features) - features:
                if f not in features:
                    n.del_feature(f)


def copy_forest(forest, features=None):
    features = set(features if features else forest[0].features)
    copied_forest = []
    for tree in forest:
        copied_tree = TreeNode()
        todo = [(tree, copied_tree)]
        copied_forest.append(copied_tree)
        while todo:
            n, copied_n = todo.pop()
            copied_n.dist = n.dist
            copied_n.support = n.support
            copied_n.name = n.name
            for f in features:
                if hasattr(n, f):
                    copied_n.add_feature(f, getattr(n, f))
            for c in n.children:
                todo.append((c, copied_n.add_child()))
    return copied_forest


def main():
    """
    Entry point, calling :py:func:`pastml.tree.name_tree` with command-line arguments.

    :return: void
    """
    import argparse

    parser = argparse.ArgumentParser(description="Name internal tree nodes as PastMl would.", prog='name_tree')

    parser.add_argument('-i', '--in_tree', help="input tree(s) in newick or nexus format (must be rooted).",
                        type=str, required=True)

    parser.add_argument('-o', '--out_tree', required=False, type=str,
                        help="path where to save the named output tree(s) in newick format.")

    parser.add_argument('--root_date', required=False, default=None,
                        help="date(s) of the root(s) (for dated tree(s) only), "
                             "if specified, the corresponding dates will be added to the tree node annotations.",
                        type=str, nargs='*')
    parser.add_argument('-c', '--columns', nargs='*',
                        help="names of the annotation columns of the input tree (if any) "
                             "to be kept in the output tree. "
                             "If there are LSD2-like date-related columns ({} and {}) present in the tree, "
                             "they will be kept in any case, so no need to specify them among columns here."
                        .format(DATE, DATE_CI),
                        type=str)
    parser.add_argument('--version', action='version',
                        version='%(prog)s {version}'.format(version=PASTML_VERSION))

    params = parser.parse_args()

    roots = read_forest(params.in_tree, columns=params.columns)
    root_dates = None
    if params.root_date is not None:
        root_dates = [parse_date(d) for d in params.root_date]
        if 1 < len(root_dates) < len(roots):
            raise ValueError('{} trees are given, but only {} root dates.'.format(len(roots), len(root_dates)))
        elif 1 == len(root_dates):
            root_dates *= len(roots)
    annotate_dates(roots, root_dates=root_dates, annotate_zeros=False)
    for i, tree in enumerate(roots):
        name_tree(tree, suffix='' if len(roots) == 1 else '_{}'.format(i))
    with open(params.out_tree, 'w+') as f:
        f.write(
            '\n'.join(
                [root.write(format_root_node=True, format=3,
                            features=[DATE, DATE_CI] + (params.columns if params.columns else []))
                 for root in roots]))


if '__main__' == __name__:
    main()



import logging

import pandas as pd
import numpy as np


def get_min_forest_stats(forest):
    len_sum = 0
    num_zero_nodes = 0
    num_tips = 0
    num_nodes = 0

    for tree in forest:
        for node in tree.traverse():
            num_nodes += 1

            if not node.dist:
                num_zero_nodes += 1

            len_sum += node.dist

            if node.is_leaf():
                num_tips += 1

    avg_len = len_sum / (num_nodes - num_zero_nodes)
    return [avg_len, num_nodes, num_tips, len_sum]


class ForestStats(object):

    def __init__(self, forest):
        self.avg_nonzero_brlen, self.num_nodes, self.num_tips, self.forest_length = get_forest_stats(forest)
        self.num_trees = len(forest)


def get_forest_stats(forest):
    len_sum_ext = 0
    len_sum_int = 0
    num_zero_nodes = 0
    max_polynomy = 0
    max_len_ext = 0
    max_len_int = 0
    min_len_ext = np.inf
    min_len_int = np.inf
    num_tips = 0
    num_nodes = 0
    num_zero_tips = 0
    tip_len_sum = 0

    for tree in forest:
        for node in tree.traverse():
            num_nodes += 1
            max_polynomy = max(len(node.children), max_polynomy)

            if not node.dist:
                num_zero_nodes += 1

            if node.is_leaf():
                num_tips += 1
                tip_len_sum += node.dist
                if node.dist:
                    min_len_ext = min(node.dist, min_len_ext)
                    len_sum_ext += node.dist
                    max_len_ext = max(max_len_ext, node.dist)
                else:
                    num_zero_tips += 1
            else:
                if node.dist:
                    min_len_int = min(node.dist, min_len_int)
                    len_sum_int += node.dist
                    max_len_int = max(max_len_int, node.dist)

    avg_len = (len_sum_ext + len_sum_int) / (num_nodes - num_zero_nodes) if num_nodes > num_zero_nodes else 0
    avg_len_ext = len_sum_ext / (num_tips - num_zero_tips) if num_tips > num_zero_tips else 0
    avg_len_int = len_sum_int / (num_nodes - num_tips - num_zero_nodes + num_zero_tips) \
        if (num_nodes - num_tips - num_zero_nodes + num_zero_tips) > 0 else 0

    logging.getLogger('pastml').debug('\n=============TREE STATISTICS===================\n'
                                      '\tnumber of tips:\t{}\n'
                                      '\tnumber of zero-branch tips:\t{}\n'
                                      '\tnumber of internal nodes:\t{}\n'
                                      '\tmax number of children per node:\t{}\n'
                                      '\tmax tip branch length:\t{:.5f}\n'
                                      '\tmax internal branch length:\t{:.5f}\n'
                                      '\tmin non-zero tip branch length:\t{:.5f}\n'
                                      '\tmin non-zero internal branch length:\t{:.5f}\n'
                                      '\tavg non-zero tip branch length:\t{:.5f}\n'
                                      '\tavg non-zero internal branch length:\t{:.5f}\n'
                                      '\tavg non-zero branch length:\t{:.5f}.'
                                      .format(num_tips,
                                              num_zero_tips,
                                              num_nodes - num_tips,
                                              max_polynomy,
                                              max_len_ext,
                                              max_len_int,
                                              min_len_ext,
                                              min_len_int,
                                              avg_len_ext,
                                              avg_len_int,
                                              avg_len))
    return [avg_len, num_nodes, num_tips, len_sum_ext + len_sum_int]


def df2gdf(df):
    df.fillna('', inplace=True)
    gb = df.groupby(df.index)
    gdf = pd.DataFrame(columns=df.columns)
    for c in df.columns:
        gdf[c] = gb[c].apply(lambda vs: {v for v in vs if not pd.isnull(v) and v != ''})
    return gdf


def preannotate_forest(forest, df=None, gdf=None):
    if gdf is None:
        gdf = df2gdf(df)
    for tree in forest:
        for node in tree.traverse('postorder'):
            if node.name in gdf.index:
                node.add_features(**gdf.loc[node.name, :].to_dict())
            else:
                for c in gdf.columns:
                    node.del_feature(c)
    return gdf.columns, gdf



import logging
import os

import numpy as np
import pandas as pd

from pastml import NUM_NODES, NUM_TIPS


MODEL = 'model'
CHANGES_PER_AVG_BRANCH = 'state_changes_per_avg_branch'
SCALING_FACTOR = 'scaling_factor'
SMOOTHING_FACTOR = 'smoothing_factor'
FREQUENCIES = 'frequencies'


class Model(object):

    def __init__(self, states, forest_stats, sf=None, tau=0, optimise_tau=False,
                 parameter_file=None, reoptimise=False, **kwargs):
        self._name = None
        self._states = np.sort(states)
        self._forest_stats = forest_stats
        self._optimise_tau = optimise_tau
        self._optimise_sf = True
        self._sf = None
        self._tau = None
        self.parse_parameters(parameter_file, reoptimise)
        if self._sf is None:
            self._sf = sf if sf is not None else 1. / forest_stats.avg_nonzero_brlen
        if self._tau is None:
            self._tau = tau if tau else 0

        self.calc_tau_factor()

        self._extra_params_fixed = False

    def calc_tau_factor(self):
        self._tau_factor = \
            self._forest_stats.forest_length / (self._forest_stats.forest_length
                                                + self._tau * (self._forest_stats.num_nodes - 1)) if self._tau else 1

    def __str__(self):
        return \
            'Model {} with parameter values:\n' \
            '{}'.format(self.name,
                        self._print_parameters())

    def _print_parameters(self):
        """
        Constructs a string representing parameter values (to be used to logging).

        :return: str representing parameter values
        """
        return self._print_basic_parameters()

    def _print_basic_parameters(self):
        return '\tscaling factor:\t{:.6f}, i.e. {:.6f} changes per avg branch\t{}\n' \
               '\tsmoothing factor:\t{:.6f}\t{}\n'\
            .format(self.sf, self.forest_stats.avg_nonzero_brlen * self.sf,
                    '(optimised)' if self._optimise_sf else '(fixed)',
                    self.tau, '(optimised)' if self._optimise_tau else '(fixed)')

    def save_parameters(self, filehandle):
        """
        Writes this model parameter values to the parameter file (in the same format as the input parameter file).

        :param filehandle: filehandle for the file where the parameter values should be written.
        :return: void
        """
        filehandle.write('{}\t{}\n'.format(MODEL, self.name))
        filehandle.write('{}\t{}\n'.format(NUM_NODES, self.forest_stats.num_nodes))
        filehandle.write('{}\t{}\n'.format(NUM_TIPS, self.forest_stats.num_tips))
        filehandle.write('{}\t{}\n'.format(SCALING_FACTOR, self.sf))
        filehandle.write('{}\t{}\n'.format(CHANGES_PER_AVG_BRANCH, self.sf * self.forest_stats.avg_nonzero_brlen))
        filehandle.write('{}\t{}\n'.format(SMOOTHING_FACTOR, self.tau))

    def extra_params_fixed(self):
        return self._extra_params_fixed

    def fix_extra_params(self):
        self._extra_params_fixed = True

    def unfix_extra_params(self):
        self._extra_params_fixed = False

    @property
    def forest_stats(self):
        return self._forest_stats

    @forest_stats.setter
    def forest_stats(self, forest_stats):
        self._forest_stats = forest_stats
        self.calc_tau_factor()

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def states(self):
        return self._states

    @states.setter
    def states(self, states):
        self._states = states

    @property
    def sf(self):
        return self._sf

    @sf.setter
    def sf(self, sf):
        if self._optimise_sf:
            self._sf = sf
        else:
            raise NotImplementedError('The scaling factor is preset and cannot be changed.')

    @property
    def tau(self):
        return self._tau

    @tau.setter
    def tau(self, tau):
        if self._optimise_tau:
            self._tau = tau
            self.calc_tau_factor()
        else:
            raise NotImplementedError('Tau is preset and cannot be changed.')

    def get_Pij_t(self, t, *args, **kwargs):
        """
        Returns a function for calculation of probability matrix of substitutions i->j over time t.

        :return: probability matrix
        :rtype: function
        """
        raise NotImplementedError("Please implement this method in the Model subclass")

    def set_params_from_optimised(self, ps, **kwargs):
        """
        Update this model parameter values from a vector representing parameters
        for the likelihood optimization algorithm.

        :param ps: np.array containing parameters of the likelihood optimization algorithm
        :param kwargs: dict of eventual other arguments
        :return: void, update this model
        """
        if self._optimise_sf:
            self.sf = ps[0]
        if self._optimise_tau:
            self.tau = ps[1 if self._optimise_sf else 0]

    def get_optimised_parameters(self):
        """
        Converts this model parameters to a vector representing parameters
        for the likelihood optimization algorithm.

        :return: np.array containing parameters of the likelihood optimization algorithm
        """
        return np.hstack(([self.sf] if self._optimise_sf else [],
                          [self.tau] if self._optimise_tau else []))

    def get_bounds(self):
        """
        Get bounds for parameters for likelihood optimization algorithm.

        :return: np.array containing lower and upper (potentially infinite) bounds for each parameter
        """
        bounds = []
        if self._optimise_sf:
            bounds += [np.array([0.001 / self.forest_stats.avg_nonzero_brlen,
                                 10. / self.forest_stats.avg_nonzero_brlen])]
        if self._optimise_tau:
            bounds += [np.array([0, self.forest_stats.avg_nonzero_brlen])]
        return np.array(bounds, np.float64)

    def get_num_params(self):
        """
        Returns the number of optimized parameters for this model.

        :return: the number of optimized parameters
        """
        return (1 if self._optimise_sf else 0) + (1 if self._optimise_tau else 0)

    def parse_parameters(self, params, reoptimise=False):
        """
        Update this model's values from the input parameters.
        The input might contain:
        (1) the scaling factor, by which each branch length will be multiplied (optional).
            The key for this parameter is pastml.models.SCALING_FACTOR;
        (2) the smoothing factor, which will be added to each branch length
            before the branches are renormalized to keep the initial tree length (optional).
            The key for this parameter is pastml.models.SMOOTHING_FACTOR;

        :param params: dict {key->value}
            or a path to the file containing a tab-delimited table with the first column containing keys
                and the second (named 'value') containing values.
        :param reoptimise: whether these model parameters should be treated as starting values (True)
            or as fixed values (False)
        :return: dict with parameter values
        """

        logger = logging.getLogger('pastml')
        frequencies, sf, kappa, tau = None, None, None, None
        if params is None:
            return {}
        if not isinstance(params, str) and not isinstance(params, dict):
            raise ValueError('Parameters must be specified either as a dict or as a path to a csv file, not as {}!'
                             .format(type(params)))
        if isinstance(params, str):
            if not os.path.exists(params):
                raise ValueError('The specified parameter file ({}) does not exist.'.format(params))
            try:
                param_dict = pd.read_csv(params, header=0, index_col=0, sep='\t')
                if 'value' not in param_dict.columns:
                    raise ValueError('Could not find the "value" column in the parameter file {}. '
                                     'It should be a tab-delimited file with two columns, '
                                     'the first one containing parameter names, '
                                     'and the second, named "value", containing parameter values.')
                param_dict = param_dict.to_dict()['value']
                params = param_dict
            except:
                raise ValueError('The specified parameter file {} is malformed, '
                                 'should be a tab-delimited file with two columns, '
                                 'the first one containing parameter names, '
                                 'and the second, named "value", containing parameter values.'.format(params))
        params = {str(k.encode('ASCII', 'replace').decode()): v for (k, v) in params.items()}
        if SCALING_FACTOR in params:
            self._sf = params[SCALING_FACTOR]
            try:
                self._sf = np.float64(self._sf)
                if self._sf <= 0:
                    logger.error('Scaling factor cannot be negative, ignoring the value given in parameters ({}).'
                                 .format(sf))
                    self._sf = None
                else:
                    self._optimise_sf = reoptimise
            except:
                logger.error('Scaling factor ({}) given in parameters is not float, ignoring it.'.format(sf))
                self._sf = None
        if SMOOTHING_FACTOR in params:
            self._tau = params[SMOOTHING_FACTOR]
            try:
                self._tau = np.float64(self._tau)
                if self._tau < 0:
                    logger.error(
                        'Smoothing factor cannot be negative, ignoring the value given in parameters ({}).'.format(tau))
                    self._tau = None
            except:
                logger.error('Smoothing factor ({}) given in parameters is not float, ignoring it.'.format(tau))
                self._tau = None
        return params

    def freeze(self):
        """
        Prohibit parameter optimization by setting all optimization flags to False.

        :return: void
        """
        self._optimise_sf = False
        self._optimise_tau = False

    def transform_t(self, t):
        return (t + self.tau) * self._tau_factor * self.sf

    def basic_params_fixed(self):
        return not self._optimise_tau and not self._optimise_sf


class ModelWithFrequencies(Model):

    def __init__(self, states, forest_stats, sf=None, frequencies=None, tau=0,
                 optimise_tau=False, frequency_smoothing=False, parameter_file=None, reoptimise=False, **kwargs):
        self._frequencies = None
        self._optimise_frequencies = not frequency_smoothing
        self._frequency_smoothing = frequency_smoothing
        Model.__init__(self, states, forest_stats=forest_stats,
                       sf=sf, tau=tau, optimise_tau=optimise_tau, reoptimise=reoptimise,
                       parameter_file=parameter_file, **kwargs)
        if self._frequencies is None:
            self._frequencies = frequencies if frequencies is not None \
                else np.ones(len(states), dtype=np.float64) / len(states)

    @property
    def frequencies(self):
        return self._frequencies

    @frequencies.setter
    def frequencies(self, frequencies):
        if self._optimise_frequencies or self._frequency_smoothing:
            self._frequencies = frequencies
        else:
            raise NotImplementedError('The frequencies are preset and cannot be changed.')

    def get_num_params(self):
        """
        Returns the number of optimized parameters for this model.

        :return: the number of optimized parameters
        """
        return Model.get_num_params(self) \
               + ((len(self.frequencies) - 1)
                  if self._optimise_frequencies else (1 if self._frequency_smoothing else 0))

    def set_params_from_optimised(self, ps, **kwargs):
        """
        Update this model parameter values from a vector representing parameters
        for the likelihood optimization algorithm.

        :param ps: np.array containing parameters of the likelihood optimization algorithm
        :param kwargs: dict of eventual other arguments
        :return: void, update this model
        """
        Model.set_params_from_optimised(self, ps, **kwargs)
        if not self.extra_params_fixed():
            n_freq = len(self.frequencies)
            n_params = Model.get_num_params(self)

            freqs = self.frequencies

            if self._optimise_frequencies:
                freqs = np.hstack((ps[n_params: n_params + (n_freq - 1)], [1.]))
                freqs /= freqs.sum()
                self.frequencies = freqs
            elif self._frequency_smoothing:
                freqs = freqs * self.forest_stats.num_tips
                freqs += ps[n_params]
                freqs /= freqs.sum()
                self.frequencies = freqs

    def get_optimised_parameters(self):
        """
        Converts this model parameters to a vector representing parameters
        for the likelihood optimization algorithm.

        :return: np.array containing parameters of the likelihood optimization algorithm
        """
        if not self.extra_params_fixed():
            return np.hstack((Model.get_optimised_parameters(self),
                              self.frequencies[:-1] / self.frequencies[-1] if self._optimise_frequencies
                              else ([0] if self._frequency_smoothing else [])))
        return Model.get_optimised_parameters(self)

    def get_bounds(self):
        """
        Get bounds for parameters for likelihood optimization algorithm.

        :return: np.array containing lower and upper (potentially infinite) bounds for each parameter
        """
        if not self.extra_params_fixed():
            extras = []
            if self._optimise_frequencies:
                extras += [np.array([1e-6, 10e6], np.float64)] * (len(self.frequencies) - 1)
            if self._frequency_smoothing:
                extras.append(np.array([0, self.forest_stats.num_nodes]))
            return np.array((*Model.get_bounds(self), *extras))
        return Model.get_bounds(self)

    def parse_parameters(self, params, reoptimise=False):
        """
        Update this model's values from the input parameters.
        For a model with frequencies, apart from the basic parameters
        (scaling factor and smoothing factor, see pastml.models.Model),
        the input might contain the frequency values:
        the key for each frequency value is the name of the corresponding character state.

        :param params: dict {key->value}
        :param reoptimise: whether these model parameters should be treated as starting values (True)
            or as fixed values (False)
        :return: dict with parameter values (same as input)
        """
        params = Model.parse_parameters(self, params, reoptimise)
        logger = logging.getLogger('pastml')
        known_freq_states = set(self.states) & set(params.keys())
        if known_freq_states:
            unknown_freq_states = [state for state in self.states if state not in params.keys()]
            if unknown_freq_states and not reoptimise:
                logger.error('Frequencies for some of the states ({}) are missing, '
                             'ignoring the specified frequencies.'.format(', '.join(unknown_freq_states)))
            else:
                self._frequencies = np.array([params[state] if state in params.keys() else 0 for state in self.states])
                try:
                    self._frequencies = self._frequencies.astype(np.float64)
                    if np.round(self._frequencies.sum() - 1, 2) != 0 and not reoptimise:
                        logger.error('Frequencies given in parameters ({}) do not sum up to one ({}),'
                                     'ignoring them.'.format(self._frequencies, self._frequencies.sum()))
                        self._frequencies = None
                    else:
                        if np.any(self._frequencies < 0) and not reoptimise:
                            logger.error('Some of the frequencies given in parameters ({}) are negative,'
                                         'ignoring them.'.format(self._frequencies))
                            self._frequencies = None
                        else:
                            min_freq = \
                                min(1 / self.forest_stats.num_tips,
                                    min(float(params[state]) for state in known_freq_states
                                        if float(params[state]) > 0)) / 2
                            if unknown_freq_states:
                                logger.error('Frequencies for some of the states ({}) are missing from parameters, '
                                             'setting them to {}.'.format(', '.join(unknown_freq_states), min_freq))
                            frequencies = np.maximum(self._frequencies, min_freq)
                            frequencies /= frequencies.sum()
                            self._optimise_frequencies = reoptimise and not self._frequency_smoothing
                except:
                    logger.error('Could not convert the frequencies given in parameters ({}) to float, '
                                 'ignoring them.'.format(self._frequencies))
                    self._frequencies = None
        return params

    def _print_parameters(self):
        """
        Constructs a string representing parameter values (to be used to logging).

        :return: str representing parameter values
        """
        return '{}' \
               '\tfrequencies\t{}\n' \
               '{}\n'.format(Model._print_parameters(self),
                             '(optimised)' if self._optimise_frequencies
                             else '(smoothed)' if self._frequency_smoothing else '(fixed)',
                             '\n'.join('\t\t{}:\t{:g}'.format(state, freq)
                                       for (state, freq) in zip(self.states, self.frequencies)))

    def freeze(self):
        """
        Prohibit parameter optimization by setting all optimization flags to False.

        :return: void
        """
        Model.freeze(self)
        self._optimise_frequencies = False
        self._frequency_smoothing = False

    def extra_params_fixed(self):
        return self._extra_params_fixed or Model.get_num_params(self) == self.get_num_params()

    def basic_params_fixed(self):
        return not Model.get_num_params(self)

    def save_parameters(self, filehandle):
        """
        Writes this model parameter values to the parameter file (in the same format as the input parameter file).

        :param filehandle: filehandle for the file where the parameter values should be written.
        :return: void
        """
        Model.save_parameters(self, filehandle)
        for state, frequency in zip(self.states, self.frequencies):
            filehandle.write('{}\t{}\n'.format(state, frequency))



import logging

import numpy as np

from pastml.models import ModelWithFrequencies
from pastml.models.generator import get_diagonalisation, get_pij_matrix

CUSTOM_RATES = 'CUSTOM_RATES'


def load_custom_rates(infile):
    rate_matrix = np.loadtxt(infile, dtype=np.float64, comments='#', delimiter=' ')
    if not len(rate_matrix.shape) == 2 or not rate_matrix.shape[0] == rate_matrix.shape[1]:
        raise ValueError('The input rate matrix must be squared, but yours is {}.'.format('x'.join(rate_matrix.shape)))
    if not np.all(rate_matrix == rate_matrix.transpose()):
        raise ValueError('The input rate matrix must be symmetric, but yours is not.')
    np.fill_diagonal(rate_matrix, 0)
    n = len(rate_matrix)
    if np.count_nonzero(rate_matrix) != n * (n - 1):
        logging.getLogger('pastml').warning('The rate matrix contains zero rates (apart from the diagonal).')
    with open(infile, 'r') as f:
        states = f.readlines()[0]
        if not states.startswith('#'):
            raise ValueError('The rate matrix file should start with state names, '
                             'separated by whitespaces and preceded by # .')
        states = np.array(states.strip('#').strip('\n').strip().split(' '), dtype=str)
        if len(states) != n:
            raise ValueError(
                'The number of specified state names ({}) does not correspond to the rate matrix dimensions ({}x{}).'
                .format(len(states), *rate_matrix.shape))
    new_order = np.argsort(states)
    return states[new_order], np.array(rate_matrix)[:, new_order][new_order, :]


class CustomRatesModel(ModelWithFrequencies):

    def __init__(self, forest_stats, sf=None, frequencies=None, rate_matrix_file=None, states=None, rate_matrix=None, tau=0,
                 optimise_tau=False, frequency_smoothing=False, parameter_file=None, reoptimise=False, **kwargs):
        ModelWithFrequencies.__init__(self, states=states, forest_stats=forest_stats,
                                      sf=sf, tau=tau, frequencies=frequencies, optimise_tau=optimise_tau,
                                      frequency_smoothing=frequency_smoothing, reoptimise=reoptimise,
                                      parameter_file=parameter_file, **kwargs)
        self.name = CUSTOM_RATES
        if rate_matrix_file is None and (rate_matrix is None or states is None):
            raise ValueError('Either the rate matrix file '
                             'or the rate matrix plus the states must be specified for {} model'.format(CUSTOM_RATES))
        if rate_matrix_file is None:
            # States are already set in the super constructor
            self._rate_matrix = rate_matrix
        else:
            self._states, self._rate_matrix = load_custom_rates(rate_matrix_file)
        self.D_DIAGONAL, self.A, self.A_INV = get_diagonalisation(self.frequencies, self._rate_matrix)

    @property
    def rate_matrix(self):
        return self._rate_matrix

    @rate_matrix.setter
    def rate_matrix(self, rate_matrix):
        raise NotImplementedError('The rate matrix is preset and cannot be changed.')

    @ModelWithFrequencies.frequencies.setter
    def frequencies(self, frequencies):
        if self._optimise_frequencies or self._frequency_smoothing:
            self._frequencies = frequencies
        else:
            raise NotImplementedError('The frequencies are preset and cannot be changed.')
        self.D_DIAGONAL, self.A, self.A_INV = get_diagonalisation(frequencies, self.rate_matrix)

    def get_Pij_t(self, t, *args, **kwargs):
        """
        Returns a function of t that calculates the probability matrix of substitutions i->j over time t,
        with the given rate matrix.

        :return: a function of t that calculates the probability matrix of substitutions i->j over time t.
        :rtype: lambda t: np.array
        """
        return get_pij_matrix(self.transform_t(t), self.D_DIAGONAL, self.A, self.A_INV)




from collections import Counter

import pandas as pd

from pastml.annotation import preannotate_forest
from pastml.tree import read_tree


def state_combinations(node, columns, i=0):
    if i < len(columns):
        if i == len(columns) - 1:
            for state in getattr(node, columns[i]):
                yield (state,)
        else:
            for cmb in state_combinations(node, columns, i + 1):
                for state in getattr(node, columns[i]):
                    yield (state, *cmb)


if '__main__' == __name__:
    import argparse

    parser = argparse.ArgumentParser()

    parser.add_argument('--tree', required=True, type=str)
    parser.add_argument('--acr', required=True, type=str)
    parser.add_argument('--columns', default=None, type=str, nargs='*')
    parser.add_argument('--out_log', required=True, type=str)
    params = parser.parse_args()

    df = pd.read_csv(params.acr, header=0, index_col=0, sep='\t')
    df.index = df.index.map(str)
    tree = read_tree(params.tree)
    preannotate_forest(df=df, forest=[tree])

    if params.columns is None or len(params.columns) == 0:
        params.columns = df.columns

    for n in tree.traverse():
        n.add_feature('state', set(state_combinations(n, params.columns)))
    
    from_to_count = Counter()
    for n in tree.traverse('preorder'):
        if n.is_root():
            continue
        states = getattr(n, 'state')
        up_states = getattr(n.up, 'state')
        for s in states:
            for u in up_states:
                if s != u:
                    from_to_count[(u, s)] += 1 / len(states) / len(up_states)

    df = pd.DataFrame(index=range(len(from_to_count)))
    df['from_to'] = list(from_to_count.keys())
    df['from'] = df['from_to'].apply(lambda _: _[0])
    df['to'] = df['from_to'].apply(lambda _: _[1])
    df.drop(labels=['from_to'], axis=1, inplace=True)
    df['count'] = from_to_count.values()
    df['normalized count'] = df['count'] / df['count'].sum()
    df.sort_values(by=['count', 'from', 'to'], axis=0, inplace=True, ascending=False)
    df.to_csv(params.out_log, index=False, sep='\t')
